import pytest

from nasap_net.models import Assembly, AuxEdge, BindingSite, Bond, Component
from nasap_net.reaction_exploration_im import Reaction
from nasap_net.reaction_exploration_im.lib import ReactionOutOfScopeError, \
    ReactionResolver


@pytest.fixture
def M_square() -> Component:
    return Component(
        kind='M', sites=[0, 1, 2, 3],
        aux_edges=[AuxEdge(0, 1), AuxEdge(1, 2), AuxEdge(2, 3), AuxEdge(3, 0)])

@pytest.fixture
def M_linear() -> Component:
    return Component(
        kind='M', sites=[0, 1],
        aux_edges=[AuxEdge(0, 1)])

@pytest.fixture
def L() -> Component:
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X() -> Component:
    return Component(kind='X', sites=[0])

@pytest.fixture
def MX2(M_linear, X) -> Assembly:
    """X0(0)-(0)M0(1)-(0)X1"""
    return Assembly(
        id_='MX2',
        components={'X0': X, 'M0': M_linear, 'X1': X},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]
    )

@pytest.fixture
def free_L(L) -> Assembly:
    return Assembly(id_='free_L', components={'L0': L}, bonds=[])

@pytest.fixture
def MLX(M_linear, L, X) -> Assembly:
    """X0(0)-(0)M0(1)-(0)L0(1)"""
    return Assembly(
        id_='MLX',
        components={'X0': X, 'M0': M_linear, 'L0': L},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'L0', 0)
        ]
    )

@pytest.fixture
def unresolved_MLX(M_linear, L, X) -> Assembly:
    """X0#(0)-(0)M0#(1)-(0)L0#(1)"""
    return Assembly(
        components={'X0#': X, 'M0#': M_linear, 'L0#': L},
        bonds=[Bond('X0#', 0, 'M0#', 0), Bond('M0#', 1, 'L0#', 0)
        ]
    )

@pytest.fixture
def free_X(X) -> Assembly:
    return Assembly(id_='free_X', components={'X0': X}, bonds=[])

@pytest.fixture
def unresolved_free_X(X) -> Assembly:
    # No Assembly ID
    return Assembly(components={'X0#': X}, bonds=[])


def test_inter(MX2, free_L, MLX, unresolved_MLX, free_X, unresolved_free_X):
    resolver = ReactionResolver([MX2, free_L, MLX, free_X])
    reaction = Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=unresolved_MLX,
        leaving_assem=unresolved_free_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 0),
        duplicate_count=4
    )
    result = resolver.resolve(reaction)
    assert result == Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 0),
        duplicate_count=4
    )


def test_reaction_out_of_scope_error(
        MX2, free_L, MLX, unresolved_MLX,
        free_X, unresolved_free_X):
    reaction = Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=MLX,
        leaving_assem=unresolved_free_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 0),
        duplicate_count=4
    )
    resolver1 = ReactionResolver([MX2, free_L, MLX])  # missing free_X
    resolver2 = ReactionResolver([MX2, free_L, free_X])  # missing MLX
    resolver3 = ReactionResolver([MX2, free_L])  # missing both
    with pytest.raises(ReactionOutOfScopeError):
        resolver1.resolve(reaction)
    with pytest.raises(ReactionOutOfScopeError):
        resolver2.resolve(reaction)
    with pytest.raises(ReactionOutOfScopeError):
        resolver3.resolve(reaction)


@pytest.fixture
def ML2X2(M_square, L, X) -> Assembly:
    """
    .. code-block::

                 X1
                (0)
                 |
                (1)
        X0(0)-(0)M0(2)-(0)L0(1)
                (3)
                 |
                (0)
                 L1
                (1)
    """
    return Assembly(
        id_='ML2X2',
        components={'M0': M_square, 'X0': X, 'X1': X, 'L0': L, 'L1': L},
        bonds=[
            Bond('M0', 0, 'X0', 0),
            Bond('M0', 1, 'X1', 0),
            Bond('M0', 2, 'L0', 0),
            Bond('M0', 3, 'L1', 0),
        ]
    )

@pytest.fixture
def ring(M_square, L, X) -> Assembly:
    """
    .. code-block::

             X1
            (0)
             |
            (1)
        /-(0)M0(2)-(0)L0(1)-/
            (3)
             |
            (0)
             L1
            (1)
    """
    return Assembly(
        id_='ring',
        components={'M0': M_square, 'X1': X, 'L0': L, 'L1': L},
        bonds=[
            Bond('M0', 0, 'L0', 1),
            Bond('M0', 1, 'X1', 0),
            Bond('M0', 2, 'L0', 0),
            Bond('M0', 3, 'L1', 0),
        ]
    )

@pytest.fixture
def unresolved_ring(M_square, L, X) -> Assembly:
    """
    .. code-block::

             X1#
            (0)
             |
            (1)
        /-(0)M0#(2)-(0)L0#(1)-/
            (3)
             |
            (0)
             L1#
            (1)
    """
    # No Assembly ID
    return Assembly(
        components={'M0#': M_square, 'X1#': X, 'L0#': L, 'L1#': L},
        bonds=[
            Bond('M0#', 0, 'L0#', 1),
            Bond('M0#', 1, 'X1#', 0),
            Bond('M0#', 2, 'L0#', 0),
            Bond('M0#', 3, 'L1#', 0),
        ]
    )


def test_intra(
        ML2X2, free_X, unresolved_free_X,
        ring, unresolved_ring):
    resolver = ReactionResolver([ML2X2, free_X, ring])
    reaction = Reaction(
        init_assem=ML2X2,
        entering_assem=None,
        product_assem=unresolved_ring,
        leaving_assem=unresolved_free_X,
        metal_bs=BindingSite('M0', 1),
        leaving_bs=BindingSite('X1', 0),
        entering_bs=BindingSite('L1', 1),
        duplicate_count=2
    )
    result = resolver.resolve(reaction)
    assert result == Reaction(
        init_assem=ML2X2,
        entering_assem=None,
        product_assem=ring,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', 1),
        leaving_bs=BindingSite('X1', 0),
        entering_bs=BindingSite('L1', 1),
        duplicate_count=2
    )



import os
from collections.abc import Hashable, Mapping
from typing import TypedDict

from nasap_net import Assembly, Component, cap_bindsites
from nasap_net.pipelines.lib import read_file, write_output


class CappingConfig(TypedDict):
    target_component_kind: str
    capping_component_kind: str
    capping_binding_site: str


# ============================================================
# Main Process
# ============================================================

def cap_bindsites_pipeline(
        assemblies_path: os.PathLike | str,
        components_path: os.PathLike | str,
        config_path: os.PathLike | str,
        output_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        verbose: bool = False
        ) -> None:
    # Input
    id_to_assembly: Mapping[Hashable, Assembly] = read_file(
        assemblies_path, verbose=verbose)
    components: Mapping[str, Component] = read_file(
        components_path, verbose=verbose)['component_kinds']
    config: CappingConfig = read_file(
        config_path, verbose=verbose)['capping_config']
    
    # Main process
    if verbose:
        print('Capping binding sites...')

    capped_assemblies = {}
    for id_, assembly in id_to_assembly.items():
        capped_assemblies[id_] = cap_bindsites(
            assembly, components, 
            config['target_component_kind'],
            config['capping_component_kind'],
            config['capping_binding_site'],
            copy=True)
        
    if verbose:
        print('Capping completed.')
    
    # Output
    write_output(
        output_path, capped_assemblies, 
        overwrite=overwrite, verbose=verbose,
        header='Capped assemblies')



from collections.abc import Hashable, Iterable
from typing import TypeVar


def validate_bondsets(
        bondsets: Iterable[Iterable[int]]
        ) -> None:
    """Validate bondsets.

    Check if bondsets contain duplicate bonds or duplicates among them.

    Parameters
    ----------
    bondsets : Iterable[Iterable[int]]
        Bondsets to validate.

    Raises
    ------
    ValueError
        If bondsets contain duplicate bonds or duplicates among them.
    """
    for bondset in bondsets:
        duplicate_bonds = _find_duplicates(bondset)
        if duplicate_bonds:
            raise ValueError(
                f'Bondset contains duplicate bonds: '
                f'bondset={sorted(bondset)}, '
                f'duplicate_bonds={sorted(duplicate_bonds)}')
    bondsets = [frozenset(bondset) for bondset in bondsets]
    duplicate_bondsets = _find_duplicates(bondsets)
    if duplicate_bondsets:
        raise ValueError(
            f'Duplicate bondset found: '
            f'{sorted(sorted(bondset) for bondset in duplicate_bondsets)}')


T = TypeVar('T')

def _find_duplicates(items: Iterable[T]) -> list[T]:
    duplicates: set[T] | list[T]
    seen: set[T] | list[T]

    if isinstance(next(iter(items)), Hashable):
        duplicates = set()
        seen = set()
        add_dup = duplicates.add
        add_seen = seen.add
    else:
        duplicates = []
        seen = []
        add_dup = duplicates.append
        add_seen = seen.append

    for item in items:
        if item in seen:
            add_dup(item)
        add_seen(item)

    return list(duplicates)



from collections.abc import Hashable, Mapping
from typing import TypeVar

from .symmetry_application import apply_symmetry_operation

__all__ = ['is_new_under_symmetry']

_T = TypeVar('_T', bound=Hashable)


def is_new_under_symmetry(
        found_assems: set[frozenset[_T]],
        new_assem: set[_T],
        sym_ops: Mapping[str, Mapping[_T, _T]] | None = None
        ) -> bool:
    """Check if a new assembly is not symmetry-equivalent to the found ones.

    Identity operation 'E' is not necessary to be included in sym_ops,
    because the existence of the assembly itself is checked first.
    """
    found_assems = set(found_assems)

    # Existence of the assembly itself is checked first.
    # This is why the sym_ops does not nessarily contain 'E'.

    # According to the Python 3.12.5 Documentation,
    # comparison of set and frozenset is based on their members,
    # e.g., set('abc') == frozenset('abc') returns True,
    # and set('abc') in set([frozenset('abc')]) returns True.
    # Therefore, the following line can check if the new_assem is
    # already found in found_assems even if the new_assem is set type.
    if new_assem in found_assems:
        return False
    if sym_ops is None:
        return True
    for sym_op in sym_ops.values():
        transformed_assem = apply_symmetry_operation(new_assem, sym_op)
        if transformed_assem in found_assems:
            return False
    return True



import pytest

from nasap_net import Assembly, Component, InterReaction, IntraReaction
from nasap_net.algorithms import are_equivalent_reactions


@pytest.fixture
def component_structures():
    return {
        'L': Component(['a', 'b']),
        'M': Component(['a', 'b']),
        'X': Component(['a']),
    }


@pytest.fixture
def id_to_assembly():
    return {
        # MX2: X0(a)-(a)M0(b)-(a)X1
        0: Assembly(
            {'M0': 'M', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        ),
        1: Assembly({'L0': 'L'}),  # L: (a)L0(b)
        2: Assembly({'X0': 'X'}),  # X: (a)X0
        # MLX: (a)L0(b)-(a)M0(b)-(a)X0
        3: Assembly(
            {'M0': 'M', 'L0': 'L', 'X0': 'X'},
            [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
        ),
        # ML2: (a)L0(b)-(a)M0(b)-(a)L1(b)
        4: Assembly(
            {'M0': 'M', 'L0': 'L', 'L1': 'L'},
            [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
        ),
        # M2L2X: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)
        5: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'X0': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'), 
             ('M1.b', 'L1.a')]
        ),
        # M2LX2: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)X1
        6: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'X1.a')]
        ),
        # M2L2-ring: //-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)-//
        7: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L'},
            [('M0.b', 'L0.a'), ('L0.b', 'M1.a'), ('M1.b', 'L1.a'),
             ('L1.b', 'M0.a')]
        ),
    }


def test_inter_equivalence(component_structures, id_to_assembly):
    # MX2 + L -> MLX + X
    reaction1 = InterReaction(
        init_assem_id=0,
        entering_assem_id=1,
        product_assem_id=3,
        leaving_assem_id=2,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=4,
    )
    # Same as above, but with different binding sites
    reaction2 = InterReaction(
        init_assem_id=0,
        entering_assem_id=1,
        product_assem_id=3,
        leaving_assem_id=2,
        metal_bs='M0.b',  # Different metal binding site
        leaving_bs='X1.a',  # Different leaving binding site
        entering_bs='L0.a',
        duplicate_count=4,
    )
    # Although the metal and leaving binding sites are different,
    # the pair of binding sites (metal, leaving) are equivalent, i.e.,
    # (M0.a, X0.a) is equivalent to (M0.b, X1.a).
    assert are_equivalent_reactions(
        reaction1, reaction2, id_to_assembly, component_structures)


def test_intra_equivalence(component_structures, id_to_assembly):
    # Add an additional assembly M2L3
    # M2L3: (a)L0(b)-(a)M0(b)-(a)L1(b)-(a)M1(b)-(a)L2(b)
    id_to_assembly[8] = Assembly(
        {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'L2': 'L'},
        [('L0.b', 'M0.a'), ('M0.b', 'L1.a'), ('L1.b', 'M1.a'),
         ('M1.b', 'L2.a')]
    )

    # M2L3 is needed for "the equivalent intra-molecular reactions
    # with different binding sites" described below.

    # M2L3 -> M2L2-ring + L
    reaction1 = IntraReaction(
        init_assem_id=8,
        product_assem_id=7,
        leaving_assem_id=1,
        metal_bs='M0.a',
        leaving_bs='L0.b',
        entering_bs='L2.b',
        duplicate_count=2,
    )
    # Same as above, but with different binding sites
    reaction2 = IntraReaction(
        init_assem_id=8,
        product_assem_id=7,
        leaving_assem_id=1,
        metal_bs='M1.b',  # Different metal binding site
        leaving_bs='L2.a',  # Different leaving binding site
        entering_bs='L0.a',
        duplicate_count=2,
    )
    # Although the metal and leaving binding sites are different,
    # the trio of binding sites (metal, leaving, entering) are equivalent, 
    # i.e., (M0.a, L0.b, L2.b) is equivalent to (M1.b, L2.a, L0.a).
    assert are_equivalent_reactions(
        reaction1, reaction2, id_to_assembly, component_structures)


def test_inter_intra_difference(component_structures, id_to_assembly):
    # MX2 + L -> MLX + X
    inter = InterReaction(
        init_assem_id=0,
        entering_assem_id=1,
        product_assem_id=3,
        leaving_assem_id=2,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=1,
    )
    # M2L2X -> M2L2-ring + X
    intra = IntraReaction(  # intra-molecular reaction
        init_assem_id=5,
        product_assem_id=7,
        leaving_assem_id=2,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L1.b',
        duplicate_count=1,
    )
    # IntraReaction should not be equivalent to InterReaction
    assert not are_equivalent_reactions(
        inter, intra, id_to_assembly, component_structures)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from collections import defaultdict
from collections.abc import Hashable, Mapping
from typing import TypeVar

from nasap_net import Assembly, Component, calc_graph_hash_of_assembly

__all__ = ['group_assemblies_by_hash']


_T = TypeVar('_T', bound=Hashable)

def group_assemblies_by_hash(
        id_to_assembly: Mapping[_T, Assembly],
        component_structures: Mapping[str, Component]
        ) -> dict[str, set[_T]]:
    # Group by hash
    hash_to_ids = defaultdict(set)
    for id_, assembly in id_to_assembly.items():
        hash_ = calc_graph_hash_of_assembly(assembly, component_structures)
        hash_to_ids[hash_].add(id_)
    return hash_to_ids



import os

import pytest
import yaml

from nasap_net import Assembly
from nasap_net.pipelines.bondset_to_assembly import bondsets_to_assemblies_pipeline


@pytest.fixture
def bondsets_data():
    return {
        0: [1],
        1: [1, 2],
        2: [2, 3],
        3: [1, 2, 3],
        4: [1, 2, 3, 4]
    }


@pytest.fixture
def structure_data():
    return {
        'components_and_their_kinds': {
            'M1': 'M',
            'M2': 'M',
            'L1': 'L',
            'L2': 'L',
            'L3': 'L'
        },
        'bonds_and_their_binding_sites': {
            1: ['L1.b', 'M1.a'],
            2: ['M1.b', 'L2.a'],
            3: ['L2.b', 'M2.a'],
            4: ['M2.b', 'L3.a']
        }
    }


@pytest.fixture
def expected_assemblies():
    return {
        0: Assembly({'L1': 'L', 'M1': 'M'}, [('L1.b', 'M1.a')]),
        1: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L'},
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a')]),
        2: Assembly(
            {'M1': 'M', 'L2': 'L', 'M2': 'M'},
            [('M1.b', 'L2.a'), ('L2.b', 'M2.a')]),
        3: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M'},
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a')]),
        4: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'L3': 'L'},
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'), 
             ('M2.b', 'L3.a')])
    }


def test_bondsets_to_assemblies_pipeline(
        tmp_path, bondsets_data, structure_data, expected_assemblies):
    # M2L3 linear: L-M-L-M-L
    # bonds: 1, 2, 3, 4 from left to right
    bondsets_path = tmp_path / 'bondsets.yaml'
    structure_data_path = tmp_path / 'structure_data.yaml'
    output_path = tmp_path / 'output.yaml'

    with open(bondsets_path, 'w') as f:
        yaml.dump(bondsets_data, f)
    
    with open(structure_data_path, 'w') as f:
        yaml.dump(structure_data, f)

    # Run the pipeline
    bondsets_to_assemblies_pipeline(
        bondsets_path=str(bondsets_path),
        structure_data_path=str(structure_data_path),
        output_path=str(output_path),
    )

    with open(output_path, 'r') as f:
        output_txt = f.read()

    # Check the output file
    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data == expected_assemblies


def test_bondsets_to_assemblies_pipeline_overwrite(
        tmp_path, bondsets_data, structure_data, expected_assemblies):
    bondsets_path = tmp_path / 'bondsets.yaml'
    structure_data_path = tmp_path / 'structure_data.yaml'
    output_path = tmp_path / 'output.yaml'

    with open(bondsets_path, 'w') as f:
        yaml.dump(bondsets_data, f)
    
    with open(structure_data_path, 'w') as f:
        yaml.dump(structure_data, f)
    
    # Create an initial output file
    with open(output_path, 'w') as f:
        yaml.dump({'initial': 'data'}, f)

    # Run the pipeline with overwrite
    bondsets_to_assemblies_pipeline(
        bondsets_path=str(bondsets_path),
        structure_data_path=str(structure_data_path),
        output_path=str(output_path),
        overwrite=True
    )

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)
    
    # Check that the initial data is gone
    assert 'initial' not in output_data

    assert output_data == expected_assemblies


def test_bondsets_to_assemblies_pipeline_no_overwrite(
        tmp_path, bondsets_data, structure_data):
    bondsets_path = tmp_path / 'bondsets.yaml'
    structure_data_path = tmp_path / 'structure_data.yaml'
    output_path = tmp_path / 'output.yaml'

    with open(bondsets_path, 'w') as f:
        yaml.dump(bondsets_data, f)
    
    with open(structure_data_path, 'w') as f:
        yaml.dump(structure_data, f)

    # Create an initial output file
    with open(output_path, 'w') as f:
        yaml.dump({'initial': 'data'}, f)

    # Run the pipeline without overwrite
    with pytest.raises(FileExistsError):
        bondsets_to_assemblies_pipeline(
            bondsets_path=str(bondsets_path),
            structure_data_path=str(structure_data_path),
            output_path=str(output_path),
            overwrite=False
        )

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)

    # Check that the initial data is still there
    assert 'initial' in output_data
    assert output_data['initial'] == 'data'


if __name__ == '__main__':
    pytest.main(['-v', __file__])



import pytest

from nasap_net.models import AuxEdge, Component
from nasap_net.yaml.lib import dump_components, load_components


@pytest.fixture
def components():
    return {
        'X': Component(kind='X', sites=[0]),
        'M(aux)': Component(
            kind='M(aux)', sites=[0, 1, 2],
            aux_edges=[AuxEdge(0, 1), AuxEdge(0, 2, kind='cis')]
        ),
    }

@pytest.fixture
def dumped_components():
    return """X: !Component
  kind: X
  sites: [0]
M(aux): !Component
  kind: M(aux)
  sites: [0, 1, 2]
  aux_edges:
  - sites: [0, 1]
  - sites: [0, 2]
    kind: cis
"""


def test_dump_components(components, dumped_components):
    dumped = dump_components(components)
    assert dumped == dumped_components


def test_load_components(components, dumped_components):
    loaded = load_components(dumped_components)
    assert loaded == components


def test_round_trip(components, dumped_components):
    dumped = dump_components(components)
    loaded = load_components(dumped)
    assert loaded == components



from collections.abc import Mapping

__all__ = ['update_nested_dict']


def update_nested_dict(original: Mapping, new: Mapping) -> dict:
    """Updates the first dictionary with the second dictionary.

    Nested dictionaries are updated recursively.

    Parameters
    ----------
    d1 : Mapping
        The dictionary to update.
    d2 : Mapping
        The dictionary to update with.

    Returns
    -------
    dict
        The updated dictionary.

    Examples
    --------
    >>> d1 = {'a': 1, 'b': {'c': 2, 'd': 3}}
    >>> d2 = {'b': {'c': 4}}
    >>> deep_update(d1, d2)
    {'a': 1, 'b': {'c': 4, 'd': 3}}
    """
    original = dict(original)
    for k, v in new.items():
        if isinstance(v, dict):
            original[k] = update_nested_dict(original.get(k, {}), v)
        else:
            original[k] = v
    return original



from collections.abc import Mapping

from nasap_net import Assembly, Component
from nasap_net.algorithms.aux_edge_existence import has_aux_edges
from nasap_net.algorithms.isomorphism.pure_isomorphism_check import \
    pure_is_isomorphic
from nasap_net.algorithms.isomorphism.rough_isomorphism_check import \
    is_roughly_isomorphic

__all__ = ['is_isomorphic']


def is_isomorphic(
        assem1: Assembly, assem2: Assembly,
        component_structures: Mapping[str, Component]
        ) -> bool:
    """Check if two assemblies are isomorphic."""
    if assem1.component_kinds != assem2.component_kinds:
        return False
    
    if has_aux_edges(assem1, component_structures):
        return pure_is_isomorphic(assem1, assem2, component_structures)
    else:
        return is_roughly_isomorphic(assem1, assem2)



import pytest

from nasap_net import (Assembly, InterReaction, InterReactionRich,
                       IntraReaction, IntraReactionRich,
                       embed_assemblies_into_reaction)


def test_inter():
    REACTION = InterReaction(
        0, 1, 2, 3,
        'M1.a', 'X1.a', 'L1.a', 2)

    ID_TO_ASSEMBLY = {
        0: Assembly({'M1': 'M', 'L1': 'L', 'X1': 'X'},  # MLX
                        [('M1.a', 'L1.a'), ('M1.b', 'X1.a')]),
        1: Assembly({'L1': 'L'}),  # L
        2: Assembly({'M1': 'M', 'L1': 'L', 'L2': 'L'},  # ML2
                        [('M1.a', 'L1.a'), ('M1.b', 'L2.a')]),
        3: Assembly({'X1': 'X'}),  # X
    }
        
    embed_reaction = embed_assemblies_into_reaction(
        REACTION, ID_TO_ASSEMBLY)
    
    assert embed_reaction == InterReactionRich(
        ID_TO_ASSEMBLY[0], ID_TO_ASSEMBLY[1],
        ID_TO_ASSEMBLY[2], ID_TO_ASSEMBLY[3],
        'M1.a', 'X1.a', 'L1.a', 2)


def test_intra():
    REACTION = IntraReaction(
        0, 1, 2,
        'M1.a', 'X1.a', 'L1.b', 1)

    ID_TO_ASSEMBLY = {
        0: Assembly({'M1': 'M', 'L1': 'L', 'X1': 'X'},  # MLX
                        [('M1.a', 'L1.a'), ('M1.b', 'X1.a')]),
        1: Assembly({'M1': 'M', 'L1': 'L'},  # ML
                        [('M1.a', 'L1.a'), ('M1.b', 'L1.b')]),
        2: Assembly({'X1': 'X'}),  # X
    }
    
    embed_reaction = embed_assemblies_into_reaction(
        REACTION, ID_TO_ASSEMBLY)
    
    assert embed_reaction == IntraReactionRich(
        ID_TO_ASSEMBLY[0], ID_TO_ASSEMBLY[1], ID_TO_ASSEMBLY[2],
        'M1.a', 'X1.a', 'L1.b', 1)


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest

from nasap_net import Assembly
from nasap_net.algorithms.isomorphism import rough_isomorphisms_iter
from nasap_net.utils import compare_mapping_iterables


def test_rough_isomorphisms_iter():
    ML2X2 = Assembly(
        {'M1': 'M', 'L1': 'L', 'L2': 'L', 'X1': 'X', 'X2': 'X'},
        [('M1.a', 'L1.a'), ('M1.b', 'L2.a'), ('M1.c', 'X1.a'), ('M1.d', 'X2.a')])
    
    isomorphisms = list(rough_isomorphisms_iter(ML2X2, ML2X2))

    expected = [
        {'M1': 'M1', 'L1': 'L1', 'L2': 'L2', 'X1': 'X1', 'X2': 'X2'},
        {'M1': 'M1', 'L1': 'L1', 'L2': 'L2', 'X1': 'X2', 'X2': 'X1'},
        {'M1': 'M1', 'L1': 'L2', 'L2': 'L1', 'X1': 'X1', 'X2': 'X2'},
        {'M1': 'M1', 'L1': 'L2', 'L2': 'L1', 'X1': 'X2', 'X2': 'X1'},
    ]

    assert len(isomorphisms) == 4
    assert {'M1': 'M1', 'L1': 'L1', 'L2': 'L2', 'X1': 'X1', 'X2': 'X2'} in isomorphisms
    assert {'M1': 'M1', 'L1': 'L1', 'L2': 'L2', 'X1': 'X2', 'X2': 'X1'} in isomorphisms
    assert {'M1': 'M1', 'L1': 'L2', 'L2': 'L1', 'X1': 'X1', 'X2': 'X2'} in isomorphisms
    assert {'M1': 'M1', 'L1': 'L2', 'L2': 'L1', 'X1': 'X2', 'X2': 'X1'} in isomorphisms
    assert compare_mapping_iterables(isomorphisms, expected)


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from dataclasses import dataclass, field
from functools import total_ordering

from nasap_net.types import ID, SupportsDunderLt
from .binding_site import BindingSite


@total_ordering
@dataclass(frozen=True, init=False)
class AuxEdge(SupportsDunderLt):
    """An auxiliary edge between two binding sites on the same component."""
    site_ids: frozenset[ID]
    kind: str | None = field(kw_only=True, default=None)

    def __init__(
            self, site_id1: ID, site_id2: ID, kind: str | None = None):
        if site_id1 == site_id2:
            raise ValueError("Sites in an auxiliary edge must be different.")
        object.__setattr__(
            self, 'site_ids',
            frozenset({site_id1, site_id2}))
        object.__setattr__(self, 'kind', kind)

    def __lt__(self, other):
        if not isinstance(other, AuxEdge):
            return NotImplemented
        self_values = [sorted(self.site_ids), self.kind]
        other_values = [sorted(other.site_ids), other.kind]
        return self_values < other_values

    def get_binding_sites(
            self, comp_id: ID,
    ) -> frozenset[BindingSite]:
        """Return the binding sites of this auxiliary edge."""
        site_id1, site_id2 = self.site_ids
        return frozenset({
            BindingSite(component_id=comp_id, site_id=site_id1),
            BindingSite(component_id=comp_id, site_id=site_id2)
        })



import pytest

from nasap_net.pipelines.bondset_enumeration import validate_bond_adjacency


def test_valid():
    bonds = [1, 2, 3, 4]
    bond_to_adj_bonds = {
        1: {2},
        2: {1, 3},
        3: {2, 4},
        4: {3},
    }
    # Should not raise any exception
    validate_bond_adjacency(bond_to_adj_bonds, bonds)


def test_invalid_keys():
    bonds = [1, 2, 3, 4]
    bond_to_adj_bonds = {
        1: {2},
        2: {1, 3},
        3: {2, 4},
        5: {3},  # Invalid key
    }
    with pytest.raises(
            ValueError, 
            match='Keys in "bond_adjacency" must be the same as "bonds".'):
        validate_bond_adjacency(bond_to_adj_bonds, bonds)


def test_invalid_values():
    bonds = [1, 2, 3, 4]
    bond_to_adj_bonds = {
        1: {2},
        2: {1, 3},
        3: {2, 5},  # Invalid value
        4: {3},
    }
    with pytest.raises(
            ValueError, 
            match='All elements in "bond_adjacency" must be in "bonds".'):
        validate_bond_adjacency(bond_to_adj_bonds, bonds)


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from typing import Any

import yaml

from nasap_net.models import AuxEdge, BindingSite, Component


def dump_components(components: Any) -> str:
    """Dump components to a YAML string."""
    return yaml.dump(
        components,
        Dumper=ComponentDumper,
        sort_keys=False,
        default_flow_style=None,
    )


class ComponentDumper(yaml.SafeDumper):
    def ignore_aliases(self, _):
        return True


def _component_representer(
        dumper: ComponentDumper, data: Component
) -> yaml.MappingNode:
    mapping: dict = {
        'kind': data.kind,
        'sites': sorted(data.site_ids),
    }
    if data.aux_edges:
        mapping['aux_edges'] = sorted(data.aux_edges)
    return dumper.represent_mapping('!Component', mapping)

def _binding_site_representer(
        dumper: ComponentDumper, data: BindingSite
) -> yaml.MappingNode:
    return dumper.represent_mapping('!BindingSite', {
        'component_id': data.component_id,
        'site_id': data.site_id,
    })

def _aux_edge_representer(
        dumper: ComponentDumper, data: AuxEdge
) -> yaml.MappingNode:
    mapping: dict = {'sites': list(data.site_ids)}
    if data.kind is not None:
        mapping['kind'] = data.kind
    return dumper.represent_dict(mapping)


yaml.add_representer(Component, _component_representer, Dumper=ComponentDumper)
yaml.add_representer(BindingSite, _binding_site_representer, Dumper=ComponentDumper)
yaml.add_representer(AuxEdge, _aux_edge_representer, Dumper=ComponentDumper)



import networkx as nx
import pytest

from nasap_net import Assembly, AuxEdge, Component, draw_2d


@pytest.mark.skip(reason='This is a visual test.')
def test_draw_2d():
    MLX3 = Assembly(
        {'M1': 'M', 'L1': 'L', 'X1': 'X', 'X2': 'X', 'X3': 'X'},
        [('M1.a', 'X1.a'), ('M1.b', 'X2.a'), ('M1.c', 'X3.a'), ('M1.d', 'L1.a')],
    )
    COMPONENT_STRUCTURES = {
        'M': Component(
            {'a', 'b', 'c', 'd'}, {
                AuxEdge('a', 'b', 'cis'), AuxEdge('b', 'c', 'cis'),
                AuxEdge('c', 'd', 'cis'), AuxEdge('d', 'a', 'cis')}),
        'L': Component({'a', 'b'}),
        'X': Component({'a'}),
    }

    positions = nx.spring_layout(MLX3.g_snapshot(COMPONENT_STRUCTURES))
    draw_2d(MLX3, COMPONENT_STRUCTURES, positions, node_labeling_mode='component_kind')
    draw_2d(MLX3, COMPONENT_STRUCTURES, positions, node_labeling_mode='component_id')
    draw_2d(MLX3, COMPONENT_STRUCTURES, positions, node_labeling_mode='core_and_bindsite_ids')
    draw_2d(MLX3, COMPONENT_STRUCTURES, positions, node_labeling_mode=None)
    draw_2d(
        MLX3, COMPONENT_STRUCTURES, positions, node_labeling_mode=None,
        label_aux_edges=False)


if __name__ == '__main__':
    test_draw_2d()



import pytest

from nasap_net.classes.component import check_bindsites_of_aux_edges_exists


def test_check_bindsites_of_aux_edges_exists():
    pass


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Hashable, Iterable
from typing import Generic, TypeVar, overload

import yaml

import nasap_net as rx

__all__ = ['FrozenUnorderedPair']


T = TypeVar('T', bound=Hashable)


class FrozenUnorderedPair(Generic[T], yaml.YAMLObject):
    """A frozen unordered pair of two hashable elements."""
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.Dumper
    yaml_tag = '!FrozenUnorderedPair'
    yaml_flow_style = None
    
    @overload
    def __init__(self, first: T, second: T) -> None: ...
    @overload
    def __init__(self, pair: Iterable[T]) -> None: ...
    def __init__(self, *args):
        # TODO: Add type validation.
        if len(args) == 1:
            pair = tuple(args[0])
            if len(pair) != 2:
                raise ValueError('The pair should have two elements.')
            self.__pair = pair
        elif len(args) == 2:
            self.__pair = tuple(args)
        else:
            raise ValueError('The pair should have two elements.')

    # TODO: Remove this property.
    @property
    def first(self) -> T:
        return self.__pair[0]
    
    # TODO: Remove this property.
    @property
    def second(self) -> T:
        return self.__pair[1]
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FrozenUnorderedPair):
            return False
        return set(self.__pair) == set(other.__pair)
    
    def __hash__(self) -> int:
        return hash(frozenset(self.__pair))
    
    def __repr__(self) -> str:
        return f'FrozenUnorderedPair({self.first!r}, {self.second!r})'
    
    def __iter__(self):
        return iter(self.__pair)

    @classmethod
    def from_yaml(cls, loader, node):
        data = loader.construct_sequence(node, deep=True)
        return FrozenUnorderedPair(data)
    
    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_sequence(
            cls.yaml_tag, sorted(data), flow_style=cls.yaml_flow_style)



import pytest

from nasap_net.utils import are_same_circular_perm


def test_same_permutation():
    assert are_same_circular_perm(
        [1, 2, 3, 4], [1, 2, 3, 4]) is True


def test_rotated_permutation():
    assert are_same_circular_perm(
        [1, 2, 3, 4], [3, 4, 1, 2]) is True


def test_different_permutation():
    assert are_same_circular_perm(
        [1, 2, 3, 4], [1, 2, 3, 5]) is False


def test_reverse_permutation_considered():
    assert are_same_circular_perm(
        [1, 2, 3, 4], [4, 3, 2, 1], consider_reverse=True) is True


def test_reverse_permutation_not_considered():
    assert are_same_circular_perm(
        [1, 2, 3, 4], [4, 3, 2, 1], consider_reverse=False) is False


def test_reverse_rotated_permutation_considered():
    assert are_same_circular_perm(
        [1, 2, 3, 4], [3, 2, 1, 4], consider_reverse=True) is True


def test_reverse_rotated_permutation_not_considered():
    assert are_same_circular_perm(
        [1, 2, 3, 4], [3, 2, 1, 4], consider_reverse=False) is False


def test_different_lengths():
    assert are_same_circular_perm(
        [1, 2, 3], [1, 2, 3, 4]) is False


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest

from nasap_net import Assembly, AuxEdge, Component, isomorphisms_iter


def test_isomorphisms_iter():
    ML2X2_CIS = Assembly(
        {'M1': 'M', 'L1': 'L', 'L2': 'L', 'X1': 'X', 'X2': 'X'},
        [('M1.a', 'L1.a'), ('M1.b', 'L2.a'), ('M1.c', 'X1.a'), ('M1.d', 'X2.a')])
    COMPONENT_STRUCTURES = {
        'M': Component(
            {'a', 'b', 'c', 'd'},
            {AuxEdge('a', 'b', 'cis'), AuxEdge('b', 'c', 'cis'),
             AuxEdge('c', 'd', 'cis'), AuxEdge('d', 'a', 'cis')}),
        'L': Component({'a'}),
        'X': Component({'a'})}
    
    isomorphisms = list(isomorphisms_iter(ML2X2_CIS, ML2X2_CIS, COMPONENT_STRUCTURES))

    # TODO: Refactor assertions to use compare_mapping_iterables
    assert len(isomorphisms) == 2
    assert {'M1.core': 'M1.core', 'L1.core': 'L1.core', 'L2.core': 'L2.core',
            'X1.core': 'X1.core', 'X2.core': 'X2.core',
            'M1.a': 'M1.a', 'M1.b': 'M1.b', 'M1.c': 'M1.c', 'M1.d': 'M1.d',
            'L1.a': 'L1.a', 'L2.a': 'L2.a', 'X1.a': 'X1.a', 'X2.a': 'X2.a'} in isomorphisms
    assert {'M1.core': 'M1.core', 'L1.core': 'L2.core', 'L2.core': 'L1.core',
            'X1.core': 'X2.core', 'X2.core': 'X1.core',
            'M1.a': 'M1.b', 'M1.b': 'M1.a', 'M1.c': 'M1.d', 'M1.d': 'M1.c',
            'L1.a': 'L2.a', 'L2.a': 'L1.a', 'X1.a': 'X2.a', 'X2.a': 'X1.a'} in isomorphisms


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from nasap_net.algorithms.isomorphism.isomorphism_check import *
from nasap_net.algorithms.isomorphism.iteration import *
from nasap_net.algorithms.isomorphism.pure_isomorphism_check import *
from nasap_net.algorithms.isomorphism.rough_isomorphism_check import *
from nasap_net.algorithms.isomorphism.rough_iteration import *



import itertools
from collections import defaultdict
from collections.abc import Iterator, Mapping
from functools import wraps

from nasap_net import (Assembly, Component, InterReaction, IntraReaction,
                       calc_graph_hash_of_assembly)
from nasap_net.algorithms.hashing import calc_graph_hash_of_assembly

from .inter import explore_inter_reactions
from .intra import explore_intra_reactions


def _verbose(func):
    """Decorator to add verbose logging to a function."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        verbose = kwargs.get('verbose', False)
        reaction_iterator = func(*args, **kwargs)
        for i, reaction in enumerate(reaction_iterator):
            if verbose:
                print(f'Reaction Found ({i}): {_reaction_to_str(reaction)}')
            yield reaction
    return wrapper


@_verbose
def explore_reactions(
        id_to_assembly: Mapping[int, Assembly],
        metal_kind: str, leaving_kind: str, entering_kind: str,
        component_structures: Mapping[str, Component],
        verbose: bool = False,  # Used in the decorator
        ) -> Iterator[IntraReaction | InterReaction]:
    hash_to_ids = defaultdict(list)
    for assem_id, assembly in id_to_assembly.items():
        hash_ = calc_graph_hash_of_assembly(assembly, component_structures)
        hash_to_ids[hash_].append(assem_id)
    
    # Intra-molecular ligand exchange
    for assem_id in id_to_assembly.keys():
        yield from explore_intra_reactions(
            assem_id, metal_kind, leaving_kind, entering_kind,
            id_to_assembly, hash_to_ids, component_structures)
    
    # Inter-molecular ligand exchange
    for init_assem_id, entering_assem_id in itertools.product(
            id_to_assembly.keys(), repeat=2):
        yield from explore_inter_reactions(
            init_assem_id, entering_assem_id, 
            metal_kind, leaving_kind, entering_kind,
            id_to_assembly, hash_to_ids, component_structures)


def _reaction_to_str(
        reaction: IntraReaction | InterReaction,
        ) -> str:
    init = reaction.init_assem_id
    entering = reaction.entering_assem_id
    product = reaction.product_assem_id
    leaving = reaction.leaving_assem_id
    
    left = f'{init}' if entering is None else f'{init} + {entering}'
    right = f'{product}' if leaving is None else f'{product} + {leaving}'
    equation = f'{left} -> {right}'

    metal_bs = reaction.metal_bs
    leaving_bs = reaction.leaving_bs
    entering_bs = reaction.entering_bs

    binding_sites = (
        f'metal={metal_bs}, leaving={leaving_bs}, entering={entering_bs}')
    
    dup = reaction.duplicate_count

    return f'{equation} ({binding_sites}) (x{dup})'



from typing import Literal, overload

from nasap_net import Assembly, BindsiteIdConverter

__all__ = ['cap_single_bindsite']


@overload
def cap_single_bindsite(
        assembly: Assembly, target_bindsite: str,
        cap_id: str, cap_component_kind: str, cap_bindsite: str,
        copy: Literal[True] = True
        ) -> Assembly: ...
@overload
def cap_single_bindsite(
        assembly: Assembly, target_bindsite: str,
        cap_id: str, cap_component_kind: str, cap_bindsite: str,
        copy: Literal[False]
        ) -> None: ...
def cap_single_bindsite(
        assembly: Assembly, target_bindsite: str,
        cap_id: str, cap_component_kind: str, cap_bindsite: str, 
        copy: bool = True
        ) -> Assembly | None:
    """Add a leaving ligand (cap) to the assembly."""
    id_converter = BindsiteIdConverter()
    if copy:
        assembly = assembly.deepcopy()

    assembly.add_component(cap_id, cap_component_kind)
    assembly.add_bond(id_converter.local_to_global(cap_id, cap_bindsite), target_bindsite)

    return assembly



import os

import pytest
import yaml

from nasap_net import Assembly, Component
from nasap_net.pipelines import find_unique_assemblies_pipeline


@pytest.fixture
def assemblies_data():
    return {
        0: Assembly({'L1': 'L', 'M1': 'M'}, [('L1.b', 'M1.a')]),
        1: Assembly({'M1': 'M', 'L2': 'L'}, [('M1.b', 'L2.a')]),  # Duplicate
        2: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L'},
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a')]),
        3: Assembly(
            {'M1': 'M', 'L2': 'L', 'M2': 'M'},
            [('M1.b', 'L2.a'), ('L2.b', 'M2.a')]),
        4: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M'},
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a')]),
        5: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'L3': 'L'},
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'), 
             ('M2.b', 'L3.a')])
    }

@pytest.fixture
def components_data():
    return {
        'component_kinds': {
            'L': Component(['a', 'b']),
            'M': Component(['a', 'b'])
            }}

@pytest.fixture
def expected_unique_assemblies(assemblies_data):
    return {
        0: assemblies_data[0],
        2: assemblies_data[2],
        3: assemblies_data[3],
        4: assemblies_data[4],
        5: assemblies_data[5]
    }

def test_find_unique_assemblies_pipeline(
        tmp_path, assemblies_data, components_data, 
        expected_unique_assemblies):
    assemblies_path = tmp_path / 'assemblies.yaml'
    components_path = tmp_path / 'components.yaml'
    output_path = tmp_path / 'output.yaml'

    with open(assemblies_path, 'w') as f:
        yaml.dump(assemblies_data, f)
    
    with open(components_path, 'w') as f:
        yaml.dump(components_data, f)

    # Run the pipeline
    find_unique_assemblies_pipeline(
        assemblies_path=str(assemblies_path),
        components_path=str(components_path),
        output_path=str(output_path),
    )

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data == expected_unique_assemblies


def test_find_unique_assemblies_pipeline_overwrite(
        tmp_path, assemblies_data, components_data, 
        expected_unique_assemblies):
    assemblies_path = tmp_path / 'assemblies.yaml'
    components_path = tmp_path / 'components.yaml'
    output_path = tmp_path / 'output.yaml'

    with open(assemblies_path, 'w') as f:
        yaml.dump(assemblies_data, f)
    
    with open(components_path, 'w') as f:
        yaml.dump(components_data, f)
    
    # Create an initial output file
    with open(output_path, 'w') as f:
        yaml.dump({'initial': 'data'}, f)

    # Run the pipeline with overwrite
    find_unique_assemblies_pipeline(
        assemblies_path=str(assemblies_path),
        components_path=str(components_path),
        output_path=str(output_path),
        overwrite=True
    )

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)
    
    # Check that the initial data is gone
    assert 'initial' not in output_data

    assert output_data == expected_unique_assemblies


def test_find_unique_assemblies_pipeline_no_overwrite(
        tmp_path, assemblies_data, components_data):
    assemblies_path = tmp_path / 'assemblies.yaml'
    components_path = tmp_path / 'components.yaml'
    output_path = tmp_path / 'output.yaml'

    with open(assemblies_path, 'w') as f:
        yaml.dump(assemblies_data, f)
    
    with open(components_path, 'w') as f:
        yaml.dump(components_data, f)

    # Create an initial output file
    with open(output_path, 'w') as f:
        yaml.dump({'initial': 'data'}, f)

    # Run the pipeline without overwrite
    with pytest.raises(FileExistsError):
        find_unique_assemblies_pipeline(
            assemblies_path=str(assemblies_path),
            components_path=str(components_path),
            output_path=str(output_path),
            overwrite=False
        )

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)

    # Check that the initial data is still there
    assert 'initial' in output_data
    assert output_data['initial'] == 'data'


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from dataclasses import dataclass

from nasap_net.models import Assembly, BindingSite
from nasap_net.types import ID


@dataclass(frozen=True)
class MLEKind:
    metal: str
    leaving: str
    entering: str


class DuplicationNotSetError(Exception):
    pass


@dataclass(frozen=True, init=False)
class MLE:
    metal: BindingSite
    leaving: BindingSite
    entering: BindingSite
    _duplication: int | None = None

    def __init__(
            self,
            metal: BindingSite,
            leaving: BindingSite,
            entering: BindingSite,
            *,
            duplication: int | None = None
    ) -> None:
        object.__setattr__(self, 'metal', metal)
        object.__setattr__(self, 'leaving', leaving)
        object.__setattr__(self, 'entering', entering)
        object.__setattr__(self, '_duplication', duplication)

    @property
    def duplication(self) -> int:
        if self._duplication is None:
            raise DuplicationNotSetError("Duplication count is not set.")
        return self._duplication


@dataclass(frozen=True)
class Reaction:
    init_assem: Assembly
    entering_assem: Assembly | None
    product_assem: Assembly
    leaving_assem: Assembly | None
    metal_bs: BindingSite
    leaving_bs: BindingSite
    entering_bs: BindingSite
    duplicate_count: int

    @property
    def init_assem_id(self) -> ID:
        return self.init_assem.id

    @property
    def entering_assem_id(self) -> ID | None:
        if self.entering_assem is None:
            return None
        id_ = self.entering_assem.id
        if id_ is None:
            raise ValueError("Entering assembly does not have an ID.")
        return id_



import pytest

from nasap_net.utils import resolve_chain_map


def test_basic():
    input1 = {1: 10, 2: 20}
    input2 = {10: 100, 20: 200}
    input3 = {100: 1000, 200: 2000}

    output = resolve_chain_map(input1, input2, input3)
    assert output == {1: 1000, 2: 2000}


def test_empty():
    output = resolve_chain_map()
    assert output == {}


def test_single_input():
    input1 = {1: 10}
    output = resolve_chain_map(input1)
    assert output == {1: 10}


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from collections import UserDict, defaultdict
from collections.abc import Mapping, MutableMapping
from dataclasses import dataclass
from typing import Generic

from nasap_net import Assembly, Component
from nasap_net.types import A, C, R
from ._lib import _IncorrectReactionResultError, _are_equivalent_mles, \
    _generate_sample_rev_mle
from .models import Reaction, _MLE


class IncorrectReactionResultError(ValueError):
    """
    Exception raised when the reproduced reaction result is inconsistent with
    the given result.
    """


class DuplicateReactionError(ValueError):
    """
    Exception raised when there are duplicate reactions in the input.
    """


def pair_reverse_reactions(
        id_to_reaction: Mapping[R, Reaction[A]],
        assemblies: Mapping[A, Assembly],
        components: Mapping[C, Component]
        ) -> dict[R, R | None]:
    """Pair reactions with their reverse reactions.

    Parameters
    ----------
    id_to_reaction: dict mapping reaction IDs to Reaction objects.
        All reactions must be unique. IDs can be all integers or all strings.
    assemblies: dict mapping assembly IDs to Assembly objects.
        IDs can be all integers or all strings.
    components: dict mapping component IDs to Component objects.
        IDs can be all integers or all strings.

    Returns
    -------
    dict mapping each reaction ID to its reverse reaction ID, or None if no
    reverse reaction exists.
        If reaction A is the reverse of reaction B, then the mapping will
        include both A -> B and B -> A.

    Raises
    ------
    IncorrectReactionResultError
        If the reproduced reaction result is inconsistent with the given
        result.
    DuplicateReactionError
        If there are duplicate reactions in the input.
    """
    return _pair_reverse_reactions(
        id_to_reaction, assemblies, components, skip_already_found=True)


@dataclass(frozen=True)
class _ReactionIndex(Generic[A]):
    init_assem_id: A
    entering_assem_id: A | None
    product_assem_id: A
    leaving_assem_id: A | None


class _NoOverwriteDict(UserDict):
    """A dict that raises an error when overwriting an existing key."""
    def __setitem__(self, key, value):
        if key in self.data:
            raise KeyError(f"Key '{key}' already exists.")
        super().__setitem__(key, value)


# skip_already_found option is for testing purpose.
def _pair_reverse_reactions(
        id_to_reaction: Mapping[R, Reaction[A]],
        assemblies: Mapping[A, Assembly],
        components: Mapping[C, Component],
        skip_already_found: bool = True
        ) -> dict[R, R | None]:
    """Implementation of `pair_reverse_reactions` with an option to skip
    already found reactions.
    """
    index_to_id = defaultdict(set)
    for target_reaction_id, target_reaction in id_to_reaction.items():
        index = _ReactionIndex(
            init_assem_id=target_reaction.init_assem_id,
            entering_assem_id=target_reaction.entering_assem_id,
            product_assem_id=target_reaction.product_assem_id,
            leaving_assem_id=target_reaction.leaving_assem_id
        )
        index_to_id[index].add(target_reaction_id)

    reaction_to_reverse: MutableMapping[R, R | None] = _NoOverwriteDict()

    for target_reaction_id, target_reaction in id_to_reaction.items():
        if skip_already_found and target_reaction_id in reaction_to_reverse:
            continue

        reversed_index = _ReactionIndex(
            init_assem_id=target_reaction.product_assem_id,
            entering_assem_id=target_reaction.leaving_assem_id,
            product_assem_id=target_reaction.init_assem_id,
            leaving_assem_id=target_reaction.entering_assem_id
            )

        candidate_ids = index_to_id.get(reversed_index)
        if not candidate_ids:
            reaction_to_reverse[target_reaction_id] = None
            continue

        try:
            sample_rev_mle = _generate_sample_rev_mle(
                target_reaction, assemblies, components)
        except _IncorrectReactionResultError:
            raise IncorrectReactionResultError() from None

        # Any reaction with MLE equivalent to the sample_rev_mle
        # is a reverse reaction.
        for candidate_id in candidate_ids:
            candidate = id_to_reaction[candidate_id]
            candidate_mle = _MLE(
                candidate.metal_bs, candidate.leaving_bs,
                candidate.entering_bs)

            rev_init_assembly = assemblies[target_reaction.product_assem_id]
            rev_entering_assembly = None
            if target_reaction.leaving_assem_id is not None:
                rev_entering_assembly = assemblies[
                    target_reaction.leaving_assem_id]

            if _are_equivalent_mles(
                    rev_init_assembly, rev_entering_assembly,
                    sample_rev_mle, candidate_mle,
                    components
                    ):
                reaction_to_reverse[target_reaction_id] = candidate_id
                reaction_to_reverse[candidate_id] = target_reaction_id
                # Multiple matches are impossible since there are no
                # duplicate reactions.
                break
        else:
            reaction_to_reverse[target_reaction_id] = None

    return dict(reaction_to_reverse)



import os

import pytest
import yaml

from nasap_net.pipelines.lib.saving import write_output


def test_creates_file(tmp_path):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    
    write_output(output_path, data)
    
    assert os.path.exists(output_path)
    with open(output_path, 'r') as f:
        saved_data = yaml.safe_load(f)
    assert saved_data == data


def test_overwrite_false(tmp_path):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    
    write_output(output_path, data)
    
    with pytest.raises(FileExistsError):
        write_output(output_path, data, overwrite=False)


def test_overwrite_true(tmp_path):
    output_path = tmp_path / "output.yaml"
    data1 = {"key": "value1"}
    data2 = {"key": "value2"}
    
    write_output(output_path, data1)
    write_output(output_path, data2, overwrite=True)
    
    with open(output_path, 'r') as f:
        saved_data = yaml.safe_load(f)
    assert saved_data == data2


def test_default_flow_style_true(tmp_path):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    
    write_output(output_path, data, default_flow_style=True)
    
    with open(output_path, 'r') as f:
        content = f.read()
    assert "{key: value}" in content


def test_default_flow_style_false(tmp_path):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    
    write_output(output_path, data, default_flow_style=False)
    
    with open(output_path, 'r') as f:
        content = f.read()
    assert "key: value" in content


def test_verbose_true(tmp_path, capsys):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    
    write_output(output_path, data, verbose=True)
    
    captured = capsys.readouterr()
    assert f'Saving the results to "{output_path}"...' in captured.out


def test_verbose_false(tmp_path, capsys):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    
    write_output(output_path, data, verbose=False)
    
    captured = capsys.readouterr()
    assert f'Saving the results to "{output_path}"...' not in captured.out


def test_write_output_with_filename_only(tmp_path):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    
    write_output(str(output_path), data)
    
    assert os.path.exists(output_path)
    with open(output_path, 'r') as f:
        saved_data = yaml.safe_load(f)
    assert saved_data == data


def test_header_with_verbose(tmp_path, capsys):
    output_path = tmp_path / "output.yaml"
    data = {"key": "value"}
    header = "Test Header"
    
    write_output(output_path, data, verbose=True, header=header)
    
    captured = capsys.readouterr()
    assert f'{header}: Successfully saved to "{output_path}".' in captured.out


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from __future__ import annotations

from collections.abc import Hashable, Iterable, Mapping
from typing import TypeVar

from .lib import enum_multi_bond_subsets, enum_single_bond_subsets

__all__ = ['enum_bond_subsets']

_T = TypeVar('_T', bound=Hashable)


def enum_bond_subsets(
        bonds: Iterable[_T],
        bond_to_adj_bonds: Mapping[_T, Iterable[_T]], 
        sym_ops: Mapping[str, Mapping[_T, _T]] | None = None
        ) -> set[frozenset[_T]]:
    """Enumerate connected subsets of bonds excluding symmetry-equivalent ones.

    When an assembly is represented as a set of bonds, subsets of 
    the bonds can be considered as fragments of the assembly.
    Considering the M2L3 linear assembly, L-M-L-M-L, with bonds named
    1, 2, 3, and 4 from left to right, for example, the assembly is 
    represented as {1, 2, 3, 4}, and its subsets such as {1}, {1, 2}, 
    and {2, 3} can be considered as fragments of the assembly, i.e.,
    L-M, L-M-L, and M-L-M, respectively.
    
    In this function, such subsets of bonds are enumerated
    under the following conditions:

    - Only connected subsets are enumerated.
    In the previous example, {1, 3} is not considered as a fragment.

    - Symmetry-equivalent fragments are excluded.
    If two fragments are superimposable under at least one symmetry 
    operation provided as `sym_ops`, only one of them is included 
    in the result.
    In the previous example, {1, 2} and {3, 4} are symmetry-equivalent
    under the C2 operation, thus only one of them is included in the
    result.

    Note that the result includes the assembly itself as a fragment,
    i.e., {1, 2, 3, 4} in the previous example.

    In summary, this function enumerates *connected* subsets of bonds
    *excluding symmetry-equivalent ones*.

    Parameters
    ----------
    bonds : Iterable[_T]
        An iterable of bond IDs.
        The resulting bondsets are dependent on the order of the bonds.
    bond_to_adj_bonds : Mapping[_T, Iterable[_T]]
        A dictionary mapping a bond to its adjacent bonds.
        Each key is the ID of a bond, and its value is an iterable
        of IDs of adjacent bonds.
        This dictionary is used for connectivity check.
    sym_ops : Mapping[str, Mapping[_T, _T]] | None, optional
        A dictionary of symmetry operations.
        Each key is the name of a symmetry operation, and its value is
        a mapping of bond IDs to their images under the symmetry operation.
        This dictionary is used for symmetry-equivalence check.
        If not provided, no symmetry-equivalence check is performed, 
        in other words, all the fragments are included in the result
        regardless of their symmetry equivalence.
        Identity operation 'E' is not necessary to be included.

    Returns
    -------
    set[frozenset[_T]]
        A set of connected subsets of bonds excluding symmetry-equivalent
        ones. Each subset is represented as a frozenset of bond IDs.

    Notes
    -----
    Equivalent fragments which are not superimposable under symmetry
    operations are not removed during the enumeration. In the previous
    example, {1, 2} (M-L) and {2, 3} (L-M) are equivalent but not
    superimposable under any symmetry operation of the complete
    assembly, thus both of them are included in the result.

    Examples
    --------
    M2L3 linear assembly (L-M-L-M-L) with bonds 1, 2, 3, and 4
    *without* symmetry operations.
    >>> import nasap_net
    >>> BONDS = [1, 2, 3, 4]
    >>> BOND_TO_ADJ_BONDS = {1: [2], 2: [1, 3], 3: [2, 4], 4: [3]}
    >>> nasap_net.enum_bond_subsets(BONDS, BOND_TO_ADJ_BONDS)
    {
        frozenset({1}), frozenset({2}), frozenset({3}), frozenset({4}),
        frozenset({1, 2}), frozenset({2, 3}), frozenset({3, 4}),
        frozenset({1, 2, 3}), frozenset({2, 3, 4}),
        frozenset({1, 2, 3, 4})
    }

    Same assembly with symmetry operations.
    >>> SYM_OPS = {'C2': {1: 4, 2: 3, 3: 2, 4: 1}}
    >>> nasap_net.enum_bond_subsets(BONDS, BOND_TO_ADJ_BONDS, SYM_OPS)
    {
        frozenset({1}), frozenset({2}), 
        frozenset({1, 2}), frozenset({2, 3}),
        frozenset({1, 2, 3}), 
        frozenset({1, 2, 3, 4})
    }

    Triangle with three bonds, 1, 2, and 3.
    >>> BONDS = [1, 2, 3]
    >>> BOND_TO_ADJ_BONDS = {1: [2, 3], 2: [1, 3], 3: [1, 2]}
    >>> SYM_OPS = {
    ...     'C3': {1: 2, 2: 3, 3: 1}, 
    ...     'C3^2': {1: 3, 2: 1, 3: 2},
    ...     'sigma1': {1: 1, 2: 3, 3: 2},
    ...     'sigma2': {1: 2, 2: 1, 3: 3},
    ...     'sigma3': {1: 3, 2: 2, 3: 1}
    ... }
    >>> nasap_net.enum_bond_subsets(BONDS, BOND_TO_ADJ_BONDS, SYM_OPS)
    {frozenset({1}), frozenset({1, 2}), frozenset({1, 2, 3})}
    """
    found: set[frozenset[_T]] = set()

    single_bond_subsets = enum_single_bond_subsets(bonds, sym_ops)
    found.update(single_bond_subsets)

    prev_assems = single_bond_subsets
    while prev_assems:
        # Subsets are explored by adding one bond to one of the
        # previously found subsets.
        cur_found = enum_multi_bond_subsets(
            prev_assems, bond_to_adj_bonds, sym_ops)
        prev_assems = cur_found
        found.update(cur_found)

    return found



from .connected_comp_count import count_connected
from .inter_or_intra_judgement import inter_or_intra
from .shortest_cycle_search import find_shortest_cycle
from .shortest_path_search import find_shortest_path



from collections import defaultdict
from collections.abc import Hashable, Iterable
from dataclasses import dataclass, field

from frozendict import frozendict

from nasap_net.isomorphism import is_isomorphic
from nasap_net.models import Assembly


class AssemblyNotFoundError(Exception):
    """Exception raised when no isomorphic assembly is found in the search space."""
    pass


@dataclass(frozen=True, init=False)
class IsomorphicAssemblyFinder:
    """Class to find isomorphic assemblies in a search space.

    Parameters
    ----------
    search_space : Iterable[Assembly]
        The search space of assemblies to find isomorphic assemblies from.
    """
    search_space: frozenset[Assembly]
    _sig_to_assemblies: frozendict[Hashable, frozenset[Assembly]] = field(init=False)

    def __init__(self, search_space: Iterable[Assembly]) -> None:
        object.__setattr__(self, 'search_space', frozenset(search_space))
        sig_to_assems: defaultdict[Hashable, set[Assembly]] = defaultdict(set)
        for assembly in self.search_space:
            sig = light_signature(assembly)
            sig_to_assems[sig].add(assembly)
        object.__setattr__(self, '_sig_to_assemblies', frozendict(
            {k: frozenset(v) for k, v in sig_to_assems.items()}
        ))

    def find(self, target: Assembly) -> Assembly:
        """Find an isomorphic assembly in the search space.

        Parameters
        ----------
        target : Assembly
            The target assembly to find an isomorphic assembly for.

        Returns
        -------
        Assembly
            An assembly from the search space that is isomorphic to the target.

        Raises
        ------
        AssemblyNotFoundError
            If no isomorphic assembly is found in the search space.
        """
        sig = light_signature(target)
        candidates = self._sig_to_assemblies.get(sig)
        if not candidates:
            raise AssemblyNotFoundError(
                f"No isomorphic assembly found for {target}")
        for candidate in candidates:
            if is_isomorphic(target, candidate):
                return candidate
        raise AssemblyNotFoundError(
            f"No isomorphic assembly found for {target}")


def light_signature(assembly: Assembly) -> Hashable:
    """Compute a light signature of the assembly for quick filtering.

    Assemblies with different signatures are guaranteed to be non-isomorphic.
    Assemblies with the same signature may or may not be isomorphic.

    The signature consists of:
    - A sorted tuple of component kinds.
    - A sorted tuple of sorted tuples of bond component kinds.

    Parameters
    ----------
    assembly : Assembly
        The assembly to compute the signature for.

    Returns
    -------
    Hashable
        The light signature of the assembly.
    """
    return (
        # component kinds
        tuple(sorted(comp.kind for comp in assembly.components.values())),
        # bond component kinds
        tuple(sorted(
            tuple(sorted([
                assembly.get_component_kind_of_site(site1),
                assembly.get_component_kind_of_site(site2),
            ]))
            for site1, site2 in assembly.bonds
        ))
    )



from collections.abc import Hashable, Iterable, Mapping
from typing import TypeVar

from nasap_net import sort_bondsets

from .is_new_check import is_new_under_symmetry

__all__ = ['enum_multi_bond_subsets']

_T = TypeVar('_T', bound=Hashable)

def enum_multi_bond_subsets(
        prev_assems: set[frozenset[_T]],
        bond_to_adj_bonds: Mapping[_T, Iterable[_T]],
        sym_ops: Mapping[str, Mapping[_T, _T]] | None = None
        ) -> set[frozenset[_T]]:
    """Enumerate multi-bond subsets of bonds
    excluding disconnected ones and symmetry-equivalent ones.
    """
    bond_to_adj_bonds = {
        bond: set(adj_bonds) 
        for bond, adj_bonds in bond_to_adj_bonds.items()}
    
    found: set[frozenset[_T]] = set()

    # NOTE: The order of the iteration should be fixed to make the
    # result deterministic.
    for prev in sort_bondsets(prev_assems):
        # List adjacent bonds to the bonds in the previous assembly.
        adj_bonds: set[_T] = set()
        for bond in prev:
            adj_bonds.update(bond_to_adj_bonds[bond])
        # Bonds in the previous assembly should be excluded.
        adj_bonds -= prev

        # NOTE: The order of the iteration should be fixed to make 
        # the result deterministic.
        for adj_bond in sorted(adj_bonds):
            new_assem = prev | {adj_bond}
            if is_new_under_symmetry(found, set(new_assem), sym_ops):
                found.add(frozenset(new_assem))

    return found



from collections.abc import Iterable
from dataclasses import dataclass

from nasap_net.types import ID
from .aux_edge import AuxEdge
from .binding_site import BindingSite


@dataclass(frozen=True, init=False)
class Component:
    """Component"""
    kind: str
    site_ids: frozenset[ID]
    aux_edges: frozenset[AuxEdge]

    def __init__(
            self, kind: str, sites: Iterable[ID],
            aux_edges: Iterable[AuxEdge] | None = None
            ):
        object.__setattr__(self, 'kind', kind)
        object.__setattr__(self, 'site_ids', frozenset(sites))
        if aux_edges is None:
            aux_edges = frozenset()
        else:
            aux_edges = frozenset(aux_edges)
        object.__setattr__(self, 'aux_edges', aux_edges)

    def get_binding_sites(self, comp_id: ID) -> frozenset[BindingSite]:
        """Return the binding sites of this component."""
        return frozenset(
            BindingSite(component_id=comp_id, site_id=site_id)
            for site_id in self.site_ids
        )



from dataclasses import FrozenInstanceError

import pytest

from nasap_net.models import Component


def test_component():
    M = Component(kind="M", sites=["a", "b"])
    assert M.kind == "M"
    assert M.site_ids == frozenset({"a", "b"})


def test_component_immutability():
    M = Component(kind="M", sites=["a", "b"])
    with pytest.raises(FrozenInstanceError):
        M.kind = "M2"
    with pytest.raises(FrozenInstanceError):
        M.site_ids = frozenset({"c"})



import pytest
import yaml

from nasap_net.yaml.helper import split_yaml_documents


def test_simple_multi_doc():
    text = """
a: 1
---
b: 2
"""
    docs = list(split_yaml_documents(text))
    assert len(docs) == 2

    data1 = yaml.safe_load(docs[0])
    data2 = yaml.safe_load(docs[1])

    assert data1 == {"a": 1}
    assert data2 == {"b": 2}


def test_block_scalar_with_dashes():
    """Check that '---' inside a block scalar does not split the document."""
    text = """
note: |
  this is part of the same block
  ---
  and should NOT split here
---
another: doc
"""
    docs = list(split_yaml_documents(text))
    assert len(docs) == 2

    data1 = yaml.safe_load(docs[0])
    data2 = yaml.safe_load(docs[1])

    assert "this is part" in data1["note"]
    assert data2 == {"another": "doc"}

    # Naive split would incorrectly produce 3 documents
    naive_docs = [d.strip() for d in text.split('---') if d.strip()]
    assert len(naive_docs) == 3


def test_comments_and_empty_lines():
    text = """
# first document
name: test

---
# second document
value: 42
...
"""
    docs = list(split_yaml_documents(text))
    assert len(docs) == 2

    data1 = yaml.safe_load(docs[0])
    data2 = yaml.safe_load(docs[1])

    assert data1 == {"name": "test"}
    assert data2 == {"value": 42}


def test_invalid_yaml_raises():
    text = """
invalid:
  - a
  - b
  - : bad
"""
    with pytest.raises(yaml.YAMLError):
        list(split_yaml_documents(text))



import pytest

from nasap_net.models import Assembly, AuxEdge, BindingSite, Bond, Component
from nasap_net.reaction_exploration_im.explorer import InterReactionExplorer
from nasap_net.reaction_exploration_im.models import MLE, MLEKind, Reaction


@pytest.fixture
def M() -> Component:
    return Component(
        kind='M', sites=[0, 1, 2, 3],
        aux_edges=[AuxEdge(0, 1), AuxEdge(1, 2), AuxEdge(2, 3), AuxEdge(3, 0)])

@pytest.fixture
def L() -> Component:
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X() -> Component:
    return Component(kind='X', sites=[0])

@pytest.fixture
def MX2(M, X) -> Assembly:
    return Assembly(
        components={'M0': M, 'X0': X, 'X1': X},
        bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0)])

@pytest.fixture
def free_L(L) -> Assembly:
    return Assembly(components={'L0': L}, bonds=[])


def test_explore(MX2, free_L, M, L, X):
    renamed_MLX = Assembly(
        {'init_M0': M, 'entering_L0': L, 'init_X1': X},
        [Bond('init_M0', 0, 'entering_L0', 0),
         Bond('init_M0', 1, 'init_X1', 0)]
    )
    renamed_X = Assembly(
        {'init_X0': X},
        []
    )
    explorer = InterReactionExplorer(MX2, free_L, MLEKind('M', 'X', 'L'))
    reactions = list(explorer.explore())
    assert set(reactions) == {
        Reaction(
            init_assem=MX2,
            entering_assem=free_L,
            product_assem=renamed_MLX,
            leaving_assem=renamed_X,
            metal_bs=BindingSite('M0', 0),
            leaving_bs=BindingSite('X0', 0),
            entering_bs=BindingSite('L0', 0),
            duplicate_count=4
        )
    }


def test__iter_mles(MX2, free_L):
    explorer = InterReactionExplorer(MX2, free_L, MLEKind('M', 'X', 'L'))
    mles = set(explorer._iter_mles())
    assert mles == {
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 0)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L0', 0)),
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L0', 1)),
    }


def test__get_unique_mles(MX2, free_L):
    explorer = InterReactionExplorer(MX2, free_L, MLEKind('M', 'X', 'L'))
    mles = {
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 0)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L0', 0)),
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L0', 1)),
    }
    assert set(explorer._get_unique_mles(mles)) == {
        MLE(
            BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 0),
            duplication=4),
    }


def test__perform_reaction(MX2, free_L, M, L, X):
    # NOTE: Component IDs in the assemblies are renamed to avoid ID conflicts.
    renamed_MLX = Assembly(
        {'init_M0': M, 'entering_L0': L, 'init_X1': X},
        [Bond('init_M0', 0, 'entering_L0', 0),
         Bond('init_M0', 1, 'init_X1', 0)]
    )
    renamed_X = Assembly(
        {'init_X0': X},
        []
    )
    explorer = InterReactionExplorer(MX2, free_L, MLEKind('M', 'X', 'L'))
    mle_with_dup = MLE(
        BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 0),
        duplication=4)
    assert explorer._perform_reaction(mle_with_dup) == Reaction(
        init_assem=MX2,
        entering_assem=free_L,
        product_assem=renamed_MLX,
        leaving_assem=renamed_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 0),
        duplicate_count=4
    )



from .cached_self_isomorphism_iteration import *
from .entering_bindsite_enumeration import *
from .ml_pair_enumeration import *
from .mle_enumeration_for_intra import *
from .unique_bindsite_or_bindsite_sets import *



from unittest.mock import patch

import pytest

import nasap_net.algorithms.hashing as hashing
from nasap_net import Assembly, Component, calc_graph_hash_of_assembly
from nasap_net.algorithms.hashing import (_multi_graph_to_graph,
                                          calc_detailed_graph_hash,
                                          calc_rough_graph_hash)

# ==========================================
# Tests for calc_wl_hash_of_assembly
# ==========================================

def test_calc_wl_hash_of_assembly_with_aux_edges():
    assembly = Assembly()
    components = dict[str, Component]()

    with (
            patch('nasap_net.algorithms.hashing.has_aux_edges',
                  return_value=True) as mock_has_aux_edges,
            patch('nasap_net.algorithms.hashing.calc_detailed_graph_hash')\
                as mock_detailed_wl_hash,
            patch('nasap_net.algorithms.hashing.calc_rough_graph_hash')\
                as mock_rough_wl_hash):
        calc_graph_hash_of_assembly(assembly, components)
        
        mock_has_aux_edges.assert_called_once_with(assembly, components)
        mock_detailed_wl_hash.assert_called_once_with(assembly, components)
        mock_rough_wl_hash.assert_not_called()


def test_calc_wl_hash_of_assembly_without_aux_edges():
    assembly = Assembly()
    components = dict[str, Component]()

    with (
            patch('nasap_net.algorithms.hashing.has_aux_edges',
                  return_value=False) as mock_has_aux_edges,
            patch('nasap_net.algorithms.hashing.calc_detailed_graph_hash')\
                as mock_detailed_wl_hash,
            patch('nasap_net.algorithms.hashing.calc_rough_graph_hash')\
                as mock_rough_wl_hash):
        calc_graph_hash_of_assembly(assembly, components)
        
        mock_has_aux_edges.assert_called_once_with(assembly, components)
        mock_detailed_wl_hash.assert_not_called()
        mock_rough_wl_hash.assert_called_once_with(assembly)


# ==========================================
# Fixtures
# ==========================================

@pytest.fixture
def COMP_KIND_TO_OBJ():
    return {
        'M_WITHOUT_AUX': Component({'a', 'b', 'c', 'd'}),
        'M_WITH_AUX': Component(
            {'a', 'b', 'c', 'd'}, 
            [('a', 'b', 'cis'), ('b', 'c', 'cis'),
             ('c', 'd', 'cis'), ('d', 'a', 'cis')]),
        'L': Component({'a', 'b'}),
        'X': Component({'a'}),
    }

@pytest.fixture
def MX4_WITHOUT_AUX():
    return Assembly(
        {'M1': 'M_WITHOUT_AUX', 
         'X1': 'X', 'X2': 'X', 'X3': 'X', 'X4': 'X'},
        [('M1.a', 'X1.a'), ('M1.b', 'X2.a'),
         ('M1.c', 'X3.a'), ('M1.d', 'X4.a')])

@pytest.fixture
def ML2X2_CIS():
    return Assembly(
        {'M1': 'M_WITH_AUX', 
         'L1': 'L', 'L2': 'L',
         'X1': 'X', 'X2': 'X'},
        [('M1.a', 'L1.a'), ('M1.b', 'L2.a'),
         ('M1.c', 'X1.a'), ('M1.d', 'X2.a')])

@pytest.fixture
def ML2X2_TRANS():
    return Assembly(
        {'M1': 'M_WITH_AUX', 
         'L1': 'L', 'L2': 'L',
         'X1': 'X', 'X2': 'X'},
        [('M1.a', 'L1.a'), ('M1.b', 'X1.a'),
         ('M1.c', 'L2.a'), ('M1.d', 'X2.a')])


# ==========================================
# Tests for calc_detailed_wl_hash
# ==========================================

def test_detailed_hash_with_same_assemblies():
    MX2_ALPHA = Assembly({'M1': 'M', 'X1': 'X', 'X2': 'X'},
                     [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])
    MX2_BETA = Assembly({'M10': 'M', 'X10': 'X', 'X20': 'X'},
                     [('M10.a', 'X10.a'), ('M10.b', 'X20.a')])
    COMP_KIND_TO_OBJ = {
        'M': Component({'a', 'b'}),
        'X': Component({'a'}),
    }

    hash_mx2_alpha = calc_detailed_graph_hash(MX2_ALPHA, COMP_KIND_TO_OBJ)
    hash_mx2_beta = calc_detailed_graph_hash(MX2_BETA, COMP_KIND_TO_OBJ)

    # Since the graphs of MX2_ALPHA and MX2_BETA are isomorphic 
    # considering the component kinds, the hashes of them should be 
    # the same.
    assert hash_mx2_alpha == hash_mx2_beta


def test_detailed_hash_with_different_component_kinds():
    MX2 = Assembly({'M1': 'M', 'X1': 'X', 'X2': 'X'},
                   [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])
    MY2 = Assembly({'M1': 'M', 'Y1': 'Y', 'Y2': 'Y'},
                   [('M1.a', 'Y1.a'), ('M1.b', 'Y2.a')])
    COMP_KIND_TO_OBJ = {
        'M': Component({'a', 'b'}),
        'X': Component({'a'}),
        'Y': Component({'a'}),
    }

    hash_mx2 = calc_detailed_graph_hash(MX2, COMP_KIND_TO_OBJ)
    hash_my2 = calc_detailed_graph_hash(MY2, COMP_KIND_TO_OBJ)

    # Even though the graphs of MX2 and MY2 are isomorphic without
    # considering the component kinds, the hashes of them should be
    # different because the component kinds (i.e., 'X' and 'Y') are
    # different
    assert hash_mx2 != hash_my2


def test_detailed_hash_with_same_assemblies_and_same_aux_edges():
    MX2_ALPHA = Assembly({'M1': 'M', 'X1': 'X', 'X2': 'X'},
                        [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])
    MX2_BETA = Assembly({'M10': 'M', 'X10': 'X', 'X20': 'X'},
                        [('M10.a', 'X10.a'), ('M10.b', 'X20.a')])
    COMP_KIND_TO_OBJ = {
        'M': Component({'a', 'b'}, [('a', 'b', 'cis')]),
        'X': Component({'a'}),
    }

    hash_mx2_alpha = calc_detailed_graph_hash(MX2_ALPHA, COMP_KIND_TO_OBJ)
    hash_mx2_beta = calc_detailed_graph_hash(MX2_BETA, COMP_KIND_TO_OBJ)

    # Since the graphs of MX2_ALPHA and MX2_BETA are isomorphic
    # considering the component kinds and the kinds of the aux edges,
    # the hashes of them should be the same.
    assert hash_mx2_alpha == hash_mx2_beta


def test_detailed_hash_with_different_connectivities():
    # X1---X2---X3
    LINEAR = Assembly(
        {'X1': 'X', 'X2': 'X', 'X3': 'X'},
        [('X1.a', 'X2.a'), ('X2.a', 'X3.a')])
    #    X1
    #  /   \
    # X4----X3
    RING = Assembly(
        {'X1': 'X', 'X2': 'X', 'X3': 'X'},
        [('X1.b', 'X2.a'), ('X2.b', 'X3.a'), ('X3.b', 'X1.a')])

    hash_linear = calc_rough_graph_hash(LINEAR)
    hash_ring = calc_rough_graph_hash(RING)

    assert hash_linear != hash_ring


def test_detailed_hash_with_different_aux_edge_existence():
    MX2_ALPHA = Assembly({'M1': 'M_WITH_AUX', 'X1': 'X', 'X2': 'X'},
                        [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])
    MX2_BETA = Assembly({'M10': 'M_WITHOUT_AUX', 'X10': 'X', 'X20': 'X'},
                        [('M10.a', 'X10.a'), ('M10.b', 'X20.a')])
    COMP_KIND_TO_OBJ = {
        'M_WITH_AUX': Component({'a', 'b'}, [('a', 'b', 'cis')]),
        'M_WITHOUT_AUX': Component({'a', 'b'}),
        'X': Component({'a'}),
    }

    hash_mx2_alpha = calc_detailed_graph_hash(MX2_ALPHA, COMP_KIND_TO_OBJ)
    hash_mx2_beta = calc_detailed_graph_hash(MX2_BETA, COMP_KIND_TO_OBJ)

    # Even though the graphs of MX2_ALPHA and MX2_BETA are isomorphic
    # considering only the component kinds, the hashes of them should be
    # different because the existence of the aux edges are different.
    assert hash_mx2_alpha != hash_mx2_beta


def test_detailed_hash_with_different_aux_edge_kinds():
    MX2_ALPHA = Assembly({'M1': 'M_CIS', 'X1': 'X', 'X2': 'X'},
                        [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])
    MX2_BETA = Assembly({'M10': 'M_TRANS', 'X10': 'X', 'X20': 'X'},
                        [('M10.a', 'X10.a'), ('M10.b', 'X20.a')])
    COMP_KIND_TO_OBJ = {
        'M_CIS': Component({'a', 'b'}, [('a', 'b', 'cis')]),
        'M_TRANS': Component({'a', 'b'}, [('a', 'b', 'trans')]),
        'X': Component({'a'}),
    }

    hash_mx2_alpha = calc_detailed_graph_hash(MX2_ALPHA, COMP_KIND_TO_OBJ)
    hash_mx2_beta = calc_detailed_graph_hash(MX2_BETA, COMP_KIND_TO_OBJ)

    # Even though the graphs of MX2_ALPHA and MX2_BETA are isomorphic
    # considering only the component kinds, the hashes of them should be
    # different because the kinds of the aux edges are different.
    assert hash_mx2_alpha != hash_mx2_beta


def test_detailed_hash_with_different_configurations():
    COMP_ID_TO_KIND = {
        'M1': 'M', 'L1': 'L', 'L2': 'L',
        'X1': 'X', 'X2': 'X'}
    ML2X2_CIS = Assembly(
        COMP_ID_TO_KIND,
        [('M1.a', 'L1.a'), ('M1.b', 'L2.a'),
         ('M1.c', 'X1.a'), ('M1.d', 'X2.a')])
    ML2X2_TRANS = Assembly(
        COMP_ID_TO_KIND,
        [('M1.a', 'L1.a'), ('M1.b', 'X1.a'),
         ('M1.c', 'L2.a'), ('M1.d', 'X2.a')])
    COMP_KIND_TO_OBJ = {
        'M': Component(
            {'a', 'b', 'c', 'd'}, 
            [('a', 'b', 'cis'), ('b', 'c', 'cis'),
             ('c', 'd', 'cis'), ('d', 'a', 'cis')]),
        'L': Component({'a', 'b'}),
        'X': Component({'a'}),
    }
    hash_ml2x2_cis = calc_detailed_graph_hash(ML2X2_CIS, COMP_KIND_TO_OBJ)
    hash_ml2x2_trans = calc_detailed_graph_hash(ML2X2_TRANS, COMP_KIND_TO_OBJ)

    # Even though the graphs of ML2_CIS and ML_TRANS are isomorphic
    # considering only the component kinds, the hashes of them should be
    # different because the configurations of the aux edges are different.
    assert hash_ml2x2_cis != hash_ml2x2_trans


# ==========================================
# Tests for calc_rough_wl_hash
# ==========================================
def test_rough_hash_with_same_assemblies():
    MX2_ALPHA = Assembly({'M1': 'M', 'X1': 'X', 'X2': 'X'},
                     [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])
    MX2_BETA = Assembly({'M10': 'M', 'X10': 'X', 'X20': 'X'},
                     [('M10.a', 'X10.a'), ('M10.b', 'X20.a')])
    hash_mx2_alpha = calc_rough_graph_hash(MX2_ALPHA)
    hash_mx2_beta = calc_rough_graph_hash(MX2_BETA)

    # Since the rough graphs of MX2_ALPHA and MX2_BETA are isomorphic 
    # considering the component kinds, the hashes of them should be 
    # the same.
    assert hash_mx2_alpha == hash_mx2_beta


def test_rough_hash_with_different_component_kinds():
    MX2 = Assembly({'M1': 'M', 'X1': 'X', 'X2': 'X'},
                   [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])
    MY2 = Assembly({'M1': 'M', 'Y1': 'Y', 'Y2': 'Y'},
                   [('M1.a', 'Y1.a'), ('M1.b', 'Y2.a')])
    hash_mx2 = calc_rough_graph_hash(MX2)
    hash_my2 = calc_rough_graph_hash(MY2)

    # Even though the rough graphs of MX2 and MY2 are isomorphic without
    # considering the component kinds, the hashes of them should be
    # different because the component kinds (i.e., 'X' and 'Y') are
    # different
    assert hash_mx2 != hash_my2


def test_rough_hash_with_different_connectivities():
    # X1---X2---X3
    LINEAR = Assembly(
        {'X1': 'X', 'X2': 'X', 'X3': 'X'},
        [('X1.a', 'X2.a'), ('X2.a', 'X3.a')])
    #    X1
    #  /   \
    # X4----X3
    RING = Assembly(
        {'X1': 'X', 'X2': 'X', 'X3': 'X'},
        [('X1.b', 'X2.a'), ('X2.b', 'X3.a'), ('X3.b', 'X1.a')])

    hash_linear = calc_rough_graph_hash(LINEAR)
    hash_ring = calc_rough_graph_hash(RING)

    assert hash_linear != hash_ring


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Iterator

import networkx as nx

from nasap_net import Assembly
from nasap_net.algorithms.subassembly import component_induced_sub_assembly

__all__ = ['extract_connected_assemblies']


def extract_connected_assemblies(assembly: Assembly) -> Iterator[Assembly]:
    g = assembly.rough_g_snapshot
    for connected_components in nx.connected_components(g):
        yield component_induced_sub_assembly(assembly, connected_components)



from collections.abc import Iterable
from dataclasses import dataclass, field

from nasap_net.models import Assembly
from .isomorphic_assembly_search import AssemblyNotFoundError, \
    IsomorphicAssemblyFinder
from ..models import Reaction


class ReactionOutOfScopeError(Exception):
    """Exception raised when a reaction cannot be resolved to the assembly space."""
    pass


@dataclass(frozen=True, init=False)
class ReactionResolver:
    """Class to resolve reactions to a given assembly space.

    Parameters
    ----------
    assembly_space : Iterable[Assembly]
        The assembly space to resolve reactions against.

    Methods
    -------
    resolve(reaction: Reaction) -> Reaction
        Resolve a reaction to the assembly space.
    """
    assembly_space: frozenset[Assembly]
    finder: IsomorphicAssemblyFinder = field(init=False)

    def __init__(self, assembly_space: Iterable[Assembly]) -> None:
        object.__setattr__(
            self, 'assembly_space', frozenset(assembly_space))
        object.__setattr__(
            self, 'finder',
            IsomorphicAssemblyFinder(self.assembly_space))

    def resolve(self, reaction: Reaction) -> Reaction:
        """Resolve a reaction to the assembly space.

        "Resolving" a reaction means ensuring that the product and leaving
        assemblies exist in the provided assembly space, and replacing them
        with the corresponding isomorphic assemblies from the assembly space.

        Parameters
        ----------
        reaction : Reaction
            The reaction to resolve.

        Returns
        -------
        Reaction
            The resolved reaction with assemblies from the assembly space.

        Raises
        ------
        ReactionOutOfScopeError
            If the reaction cannot be resolved to the assembly space.
        """
        # Cond-1: The product assembly must exist in the provided assemblies.
        try:
            isom_product_assem = self.finder.find(reaction.product_assem)
        except AssemblyNotFoundError as e:
            raise ReactionOutOfScopeError("Product assembly not found") from e

        if reaction.leaving_assem is None:
            return Reaction(
                init_assem=reaction.init_assem,
                entering_assem=reaction.entering_assem,
                product_assem=isom_product_assem,
                leaving_assem=None,
                metal_bs=reaction.metal_bs,
                leaving_bs=reaction.leaving_bs,
                entering_bs=reaction.entering_bs,
                duplicate_count=reaction.duplicate_count
            )

        # Cond-2: The leaving assembly must exist in the provided assemblies.
        try:
            isom_leaving_assem = self.finder.find(reaction.leaving_assem)
        except AssemblyNotFoundError as e:
            raise ReactionOutOfScopeError("Leaving assembly not found") from e

        return Reaction(
            init_assem=reaction.init_assem,
            entering_assem=reaction.entering_assem,
            product_assem=isom_product_assem,
            leaving_assem=isom_leaving_assem,
            metal_bs=reaction.metal_bs,
            leaving_bs=reaction.leaving_bs,
            entering_bs=reaction.entering_bs,
            duplicate_count=reaction.duplicate_count
        )



import os
import tempfile
from collections.abc import Hashable
from pathlib import Path
from typing import TypeVar

from .assembly_list_concatenation import \
    concatenate_assemblies_without_isom_checks
from .bindsite_capping import cap_bindsites_pipeline
from .bondset_enumeration import enum_bond_subsets_pipeline
from .bondset_to_assembly import bondsets_to_assemblies_pipeline
from .duplicate_exclusion import find_unique_assemblies_pipeline


def enumerate_assemblies_pipeline(
        input_path: os.PathLike | str,
        output_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        verbose: bool = False,
        wip_dir: os.PathLike | str | None = None
        ) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        if wip_dir is None:
            wip_dir = temp_dir
        wip_dir = Path(wip_dir)

        resolved_sym_ops_path = wip_dir / 'resolved_sym_ops.yaml'
        wip1_bondsets_path = wip_dir / 'wip1_bondsets.yaml'
        wip2_assemblies_path = wip_dir / 'wip2_assemblies.yaml'
        wip3_unique_assemblies_path = wip_dir / 'wip3_unique_assemblies.yaml'
        wip4_capped_assemblies_path = wip_dir / 'wip4_capped_assemblies.yaml'

        if verbose:
            print('Enumerating assemblies...')
            
        enum_bond_subsets_pipeline(
            input_path, wip1_bondsets_path, overwrite=overwrite,
            path_to_output_resolved_sym_ops=resolved_sym_ops_path)
        bondsets_to_assemblies_pipeline(
            wip1_bondsets_path, input_path, wip2_assemblies_path,
            overwrite=overwrite)
        find_unique_assemblies_pipeline(
            wip2_assemblies_path, input_path, wip3_unique_assemblies_path,
            overwrite=overwrite)
        cap_bindsites_pipeline(
            wip3_unique_assemblies_path, input_path, input_path, wip4_capped_assemblies_path,
            overwrite=overwrite)
        concatenate_assemblies_without_isom_checks(
            [wip4_capped_assemblies_path], output_path, overwrite=overwrite)
        

        if verbose:
            print('Assembly enumeration completed.')
            print(f'Successfully saved to {output_path}.')



from math import inf

import networkx as nx

from nasap_net import Assembly


def find_shortest_cycle(assembly: Assembly) -> list | None:
    """Find one of the shortest cycles in the assembly.

    The length of cycle is the number of components in the cycle.
    For example, the cycle ['M1', 'L1', 'M2', 'L2'] has length 4.

    Which of the shortest cycles is returned is not guaranteed.

    Parameters
    ----------
    assembly : Assembly
        The assembly to find the shortest cycle in.

    Returns
    -------
    list or None
        The shortest cycle in the assembly.
        If there is no cycle, return None.
    """
    g = assembly.rough_g_snapshot
    cycles = list(nx.simple_cycles(g))
    if not cycles:
        return None
    return min(cycles, key=len)



import pytest

from nasap_net.pipelines.bondset_enumeration import sort_bond_subsets


def test_empty():
    result = sort_bond_subsets([])  # type: ignore
    assert result == []


def test_single_element():
    result = sort_bond_subsets([frozenset({1})])
    assert result == [[1]]


def test_multiple_elements():
    result = sort_bond_subsets([
        frozenset({1, 2, 3}),
        frozenset({1}),
        frozenset({2, 3}),
        frozenset({1, 2}),
    ])
    assert result == [
        [1],
        [1, 2],
        [2, 3],
        [1, 2, 3],
    ]


def test_with_strings():
    result = sort_bond_subsets([
        frozenset({'a', 'b', 'c'}),
        frozenset({'a'}),
        frozenset({'b', 'c'}),
        frozenset({'a', 'b'}),
    ])
    assert result == [
        ['a'],
        ['a', 'b'],
        ['b', 'c'],
        ['a', 'b', 'c'],
    ]


if __name__ == '__main__':
    pytest.main(['-v', __file__])



import os

import pytest
import yaml
from click.testing import CliRunner

from nasap_net import Assembly, Component
from nasap_net.cli.commands import run_concat_assembly_lists_pipeline


@pytest.fixture
def assemblies_data():
    return [
        {
            0: Assembly({'L1': 'L', 'M1': 'M'}, [('L1.b', 'M1.a')]),
            1: Assembly({'L2': 'L', 'M2': 'M'}, [('L2.b', 'M2.a')])
        },
        {
            2: Assembly({'L3': 'L', 'M3': 'M'}, [('L3.b', 'M3.a')]),
            3: Assembly({'L4': 'L', 'M4': 'M'}, [('L4.b', 'M4.a')])
        }
    ]


@pytest.fixture
def components_data():
    return {
        'component_kinds': {
            'L': Component({'a', 'b'}),
            'M': Component({'a', 'b'})
        }
    }


@pytest.fixture
def expected_concatenated_assemblies():
    return {
        0: Assembly({'L1': 'L', 'M1': 'M'}, [('L1.b', 'M1.a')]),
    }


def test_run_concat_assembly_lists_pipeline(tmp_path, assemblies_data, components_data, expected_concatenated_assemblies):
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        assemblies_paths = []
        for i, data in enumerate(assemblies_data):
            path = os.path.join(td, f'assemblies_{i}.yaml')
            with open(path, 'w') as f:
                yaml.dump(data, f)
            assemblies_paths.append(path)

        components_path = os.path.join(td, 'components.yaml')
        with open(components_path, 'w') as f:
            yaml.dump(components_data, f)

        output_path = os.path.join(td, 'output.yaml')

        result = runner.invoke(
            run_concat_assembly_lists_pipeline,
            [*assemblies_paths, components_path, output_path]
        )

        assert result.exit_code == 0
        assert os.path.exists(output_path)

        with open(output_path, 'r') as f:
            actual_output = yaml.safe_load(f)

        assert actual_output == expected_concatenated_assemblies


if __name__ == '__main__':
    pytest.main(['-v', __file__])


from collections.abc import Iterable, Iterator, Mapping

from nasap_net import Assembly

from .single_assembly import convert_bondset_to_assembly

__all__ = ['convert_bondsets_to_assemblies']


def convert_bondsets_to_assemblies(
        bondsets: Iterable[frozenset[str]],
        components: Mapping[str, str],
        bond_id_to_bindsites: dict[str, frozenset[str]]
        ) -> Iterator[Assembly]:
    """Converts the connected bonds to graphs."""

    for bondset in bondsets:
        yield convert_bondset_to_assembly(
            set(bondset), components, bond_id_to_bindsites)


import pytest

from nasap_net import Assembly, Component, assemblies_equal, perform_inter_exchange


def test_inter_exchange() -> None:
    # X1(a)--(a)M1(a)--(a)X2
    MX2 = Assembly(
        {'M1': 'M', 'X1': 'X', 'X2': 'X'},
        [('X1.a', 'M1.a'), ('M1.b', 'X2.a')])
    # (a)L1(b)
    L = Assembly({'L1': 'L'}, [])
    # expected product: MLX
    # (b)L1(a)--(a)M1(b)--(a)X2
    MLX = Assembly(
        {'M1': 'M', 'L1': 'L', 'X2': 'X'},
        [('L1.a', 'M1.a'), ('M1.b', 'X2.a')])
    MLX = MLX.rename_component_ids(
        {'M1': 'init_M1', 'L1': 'entering_L1', 'X2': 'init_X2'})
    X = Assembly({'X1': 'X'}, [])
    X = X.rename_component_ids({'X1': 'init_X1'})
    
    COMPONENT_KINDS = {
        'M': Component({'a', 'b'}),
        'L': Component({'a'}),
        'X': Component({'a'})}
    
    product, leaving = perform_inter_exchange(
        MX2, L, 'M1.a', 'X1.a', 'L1.a')

    assert assemblies_equal(product, MLX, COMPONENT_KINDS)
    assert leaving is not None
    assert assemblies_equal(leaving, X, COMPONENT_KINDS)


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Iterable, Iterator, Mapping

from nasap_net import (Assembly, Component, find_isomorphic_assembly,
                       perform_intra_exchange)
from nasap_net.classes.reaction import IntraReaction

from .lib import (compute_unique_bindsites_or_bindsite_sets,
                  enum_valid_mles_for_intra)


def explore_intra_reactions(
        init_assem_id: int,
        metal_kind: str, leaving_kind: str, entering_kind: str,
        id_to_assembly: Mapping[int, Assembly],
        hash_to_ids: Mapping[str, Iterable[int]],
        component_structures: Mapping[str, Component],
        ) -> Iterator[IntraReaction]:
    init_assem = id_to_assembly[init_assem_id]

    valid_mle_trios = enum_valid_mles_for_intra(
        init_assem_id, init_assem,
        metal_kind, leaving_kind, entering_kind,
        component_structures)

    for mle_bindsites, duplicate_count in compute_unique_bindsites_or_bindsite_sets(
            init_assem_id, init_assem, valid_mle_trios,
            component_structures):
        metal_bs, leaving_bs, entering_bs = mle_bindsites

        product, leaving = perform_intra_exchange(
            init_assem, metal_bs, leaving_bs, entering_bs)

        product_id = find_isomorphic_assembly(
            product, id_to_assembly, hash_to_ids,
            component_structures)
        if product_id is None:
            continue

        if leaving is None:
            yield IntraReaction(
                init_assem_id, product_id, None,
                metal_bs, leaving_bs, entering_bs, 
                duplicate_count)
            continue

        leaving_id = find_isomorphic_assembly(
            leaving, id_to_assembly, hash_to_ids,
            component_structures)
        if leaving_id is None:
            continue

        yield IntraReaction(
            init_assem_id, product_id, leaving_id,
            metal_bs, leaving_bs, entering_bs, 
            duplicate_count)



from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import Self

from nasap_net.exceptions import IDNotSetError
from nasap_net.types import ID
from .assembly import Assembly
from .bond import Bond


@dataclass(frozen=True, init=False)
class LightAssembly:
    components: dict[ID, str]
    bonds: frozenset[Bond]
    _id: ID | None

    def __init__(
            self,
            components: Mapping[ID, str],
            bonds: Iterable[Bond],
            id_: ID | None = None,
            ):
        object.__setattr__(self, 'components', components)
        object.__setattr__(self, 'bonds', frozenset(bonds))
        object.__setattr__(self, '_id', id_)

    @property
    def id(self) -> ID:
        if self._id is None:
            raise IDNotSetError('LightAssembly ID is not set.')
        return self._id

    @property
    def id_or_none(self) -> ID | None:
        return self._id

    @classmethod
    def from_assembly(cls, assembly: Assembly) -> Self:
        components = {
            comp_id: comp.kind
            for comp_id, comp in assembly.components.items()
        }
        return cls(
            components=components,
            bonds=assembly.bonds,
            id_=assembly.id_or_none,
        )



from typing import Any

import yaml

from nasap_net.models import Bond, LightAssembly


def dump_light_assemblies(assemblies: Any) -> str:
    """Dump light assemblies to a YAML string."""
    return yaml.dump(
        assemblies,
        Dumper=_LightAssemblyDumper,
        sort_keys=False,
        default_flow_style=None,
    )


class _LightAssemblyDumper(yaml.SafeDumper):
    def ignore_aliases(self, _):
        return True


def _light_assembly_representer(
        dumper: _LightAssemblyDumper, data: LightAssembly
) -> yaml.MappingNode:
    mapping: dict = {
        'components': dict(sorted(data.components.items())),
        'bonds': sorted(data.bonds),
    }
    if data.id_or_none is not None:
        mapping['id'] = data.id
    return dumper.represent_mapping('!LightAssembly', mapping)


def _bond_representer(
        dumper: _LightAssemblyDumper, data: Bond
) -> yaml.SequenceNode:
    site1, site2 = data.sites
    return dumper.represent_list([
        site1.component_id,
        site1.site_id,
        site2.component_id,
        site2.site_id,
    ])


yaml.add_representer(
    LightAssembly, _light_assembly_representer, Dumper=_LightAssemblyDumper
)
yaml.add_representer(Bond, _bond_representer, Dumper=_LightAssemblyDumper)



import pytest

from nasap_net.models import BindingSite, Bond


def test_bond():
    bond = Bond(comp_id1="L1", site1="a", comp_id2="M1", site2="b")
    assert bond.sites == (
        BindingSite(component_id="L1", site_id="a"),
        BindingSite(component_id="M1", site_id="b")
        )


def test_binding_site_ordering_in_bond():
    bond1 = Bond(comp_id1="L1", site1="a", comp_id2="M1", site2="b")
    bond2 = Bond(comp_id1="M1", site1="b", comp_id2="L1", site2="a")
    assert bond1 == bond2
    assert bond1.sites[0] < bond1.sites[1]


def test_bond_same_site_error():
    with pytest.raises(ValueError):
        Bond(comp_id1="M1", site1="a", comp_id2="M1", site2="b")
    # Should not raise
    Bond(comp_id1="M1", site1="a", comp_id2="M2", site2="a")



import re

__all__ = ['validate_name_of_aux_type']


def validate_name_of_aux_type(aux_type: str) -> None:
    """Check if the auxiliary type is valid.

    The auxiliary type should be a string that follows the pattern 
    '[A-Za-z0-9_-]+'.
    """
    if not re.fullmatch(r'[A-Za-z0-9_-]+', aux_type):
        raise ValueError(
            'The auxiliary type should follow the pattern [A-Za-z0-9_-]+.')



from collections.abc import Mapping

import networkx as nx
from networkx.algorithms.isomorphism import (categorical_edge_match,
                                             categorical_node_match)

from nasap_net import Assembly, Component

__all__ = ['pure_is_isomorphic']


def pure_is_isomorphic(
        assem1: Assembly, assem2: Assembly,
        component_kinds: Mapping[str, Component]
        ) -> bool:
    """Check if two assemblies are isomorphic."""
    node_match = categorical_node_match('component_kind', None)
    edge_match = categorical_edge_match('aux_kind', None)
    # QUESTION: Should we use VF2++ instead of VF2?
    return nx.is_isomorphic(
        assem1.g_snapshot(component_kinds), assem2.g_snapshot(component_kinds),
        node_match=node_match, edge_match=edge_match)



from collections.abc import Mapping

import networkx as nx

from nasap_net import Assembly, Component
from nasap_net.algorithms.aux_edge_existence import has_aux_edges


def calc_graph_hash_of_assembly(
        assembly: Assembly, 
        component_structures: Mapping[str, Component]
        ) -> str:
    """Calculate the hash of the graph of the assembly.

    Parameters
    ----------
    assembly : Assembly
        The assembly to calculate the hash of.
    component_structures : Mapping[str, Component]
        The structures of the components in the assembly.

    Returns
    -------
    str
        The hash of the graph of the assembly.

    Notes
    -----
    The hash is calculated using the Weisfeiler-Lehman graph hash
    algorithm. The hash is calculated based on the rough graph if the
    assembly does not have auxiliary edges, and based on the detailed
    graph if the assembly has auxiliary edges.
    """
    if has_aux_edges(assembly, component_structures):
        return calc_detailed_graph_hash(assembly, component_structures)
    else:
        return calc_rough_graph_hash(assembly)


def calc_rough_graph_hash(assembly: Assembly) -> str:
    # Since nx.weisfeiler_lehman_graph_hash() is not supported for
    # MultiGraph, which is used for the rough graph, we need to convert
    # it to Graph, i.e., remove parallel edges.
    # Consequently, the result of this function will be the same for
    # assemblies with the same rough graph except for the existence of
    # parallel edges.
    rough_g = assembly.rough_g_snapshot
    rough_g_without_parallel_edges = _multi_graph_to_graph(rough_g)
    return nx.weisfeiler_lehman_graph_hash(
        rough_g_without_parallel_edges, node_attr='component_kind')


def _multi_graph_to_graph(G_multi: nx.MultiGraph) -> nx.Graph:
    G_single = nx.Graph()
    for node, data in G_multi.nodes(data=True):
        G_single.add_node(node, **data)
    for u, v, data in G_multi.edges(data=True):
        G_single.add_edge(u, v)
    return G_single


def calc_detailed_graph_hash(
        assembly: Assembly, 
        component_structures: Mapping[str, Component]
        ) -> str:
    g = assembly.g_snapshot(component_structures)
    _add_attr_for_hash(g)
    
    return nx.weisfeiler_lehman_graph_hash(
        g, node_attr='for_hash', edge_attr='for_hash')


def _add_attr_for_hash(g: nx.Graph) -> None:
    for node, data in g.nodes(data=True):
        if data['core_or_bindsite'] == 'core':
            data['for_hash'] = data['component_kind']
        else:
            data['for_hash'] = None

    for u, v, data in g.edges(data=True):
        if 'aux_kind' in data:
            data['for_hash'] = data['aux_kind']
        else:
            data['for_hash'] = None



__all__ = ['BindsiteIdConverter']


class BindsiteIdConverter:
    DELIMITER = '.'

    def local_to_global(self, comp_id: str, local_bs_id: str) -> str:
        return f'{comp_id}{self.DELIMITER}{local_bs_id}'

    def global_to_local(self, abs_bs_id: str) -> tuple[str, str]:
        splited = abs_bs_id.split(self.DELIMITER)
        assert len(splited) == 2
        comp_id, local_bs_id = splited
        return comp_id, local_bs_id



from typing import TypeAlias

import pytest

from nasap_net import (Assembly, Component, InterReaction, IntraReaction,
                       explore_reactions)
from nasap_net.algorithms import are_equivalent_reaction_sets
from nasap_net.reaction_exploration import (
    clear_cache_for_compute_unique_bindsites_or_bindsite_sets,
    clear_cache_for_enum_valid_entering_bindsites,
    clear_cache_for_enum_valid_ml_pairs,
    clear_cache_for_enum_valid_mles_for_intra,
    clear_cache_for_iter_self_isomorphisms)

Reaction: TypeAlias = IntraReaction | InterReaction


@pytest.fixture(autouse=True)
def clear_caches():
    """Clear caches before each test to ensure independence."""
    clear_cache_for_compute_unique_bindsites_or_bindsite_sets()
    clear_cache_for_iter_self_isomorphisms()
    clear_cache_for_enum_valid_entering_bindsites()
    clear_cache_for_enum_valid_ml_pairs()
    clear_cache_for_enum_valid_mles_for_intra()


def test_comprehensive_reaction_sets():
    component_structures = {
        'M': Component(['a', 'b']),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
    }

    id_to_assembly = {
        # MX2: X0(a)-(a)M0(b)-(a)X1
        0: Assembly(
            {'M0': 'M', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        ),
        1: Assembly({'L0': 'L'}),  # L: (a)L0(b)
        2: Assembly({'X0': 'X'}),  # X: (a)X0
        # MLX: (a)L0(b)-(a)M0(b)-(a)X0
        3: Assembly(
            {'M0': 'M', 'L0': 'L', 'X0': 'X'},
            [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
        ),
        # ML2: (a)L0(b)-(a)M0(b)-(a)L1(b)
        4: Assembly(
            {'M0': 'M', 'L0': 'L', 'L1': 'L'},
            [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
        ),
        # M2L2X: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)
        5: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'X0': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'L1.a')]
        ),
        # M2LX2: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)X1
        6: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'X1.a')]
        ),
        # M2L2-ring: //-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)-//
        7: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L'},
            [('M0.b', 'L0.a'), ('L0.b', 'M1.a'), ('M1.b', 'L1.a'),
             ('L1.b', 'M0.a')]
        ),
    }

    # All "X to L" reactions among the assemblies in id_to_assembly
    # including intra- and inter-molecular reactions:
    expected: dict[str, Reaction] = {
        # Intra-molecular reactions
        'M2L2X -> M2L2-ring + X': IntraReaction(
            init_assem_id=5, 
            product_assem_id=7, leaving_assem_id=2, 
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L1.b',
            duplicate_count=1,
        ),
        # Inter-molecular reactions
        'MX2 + L -> MLX + X': InterReaction(
            init_assem_id=0, entering_assem_id=1,
            product_assem_id=3, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
        'MX2 + MLX -> M2LX2 + X': InterReaction(
            init_assem_id=0, entering_assem_id=3,
            product_assem_id=6, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'MX2 + ML2 -> M2L2X + X': InterReaction(
            init_assem_id=0, entering_assem_id=4,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
        'MLX + L -> ML2 + X': InterReaction(
            init_assem_id=3, entering_assem_id=1,
            product_assem_id=4, leaving_assem_id=2,
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'MLX + MLX -> M2L2X + X': InterReaction(
            init_assem_id=3, entering_assem_id=3,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'M2LX2 + L -> M2L2X + X': InterReaction(
            init_assem_id=6, entering_assem_id=1,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
    }

    result = explore_reactions(
        id_to_assembly,
        metal_kind='M', leaving_kind='X', entering_kind='L',
        component_structures=component_structures
    )

    assert are_equivalent_reaction_sets(
        result, list(expected.values()),
        id_to_assembly=id_to_assembly,
        component_structures=component_structures
    ), "The reaction sets are not equivalent."


def test_aux_edges():
    component_structures = {
        'M': Component(  # component with auxiliary edges
            ['a', 'b', 'c', 'd'], 
            [('a', 'b', 'cis'), ('b', 'c', 'cis'), 
             ('c', 'd', 'cis'), ('d', 'a', 'cis')]),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
    }

    id_to_assembly = {
        0: Assembly(  # MX4
            {'M0': 'M', 'X0': 'X', 'X1': 'X', 'X2': 'X', 'X3': 'X'},
            [('M0.a', 'X0.a'), ('M0.b', 'X1.a'), 
             ('M0.c', 'X2.a'), ('M0.d', 'X3.a')]
        ),
        1: Assembly({'L0': 'L'}),  # L
        2: Assembly({'X0': 'X'}),  # X
        3: Assembly(  # MLX3
            {'M0': 'M', 'L0': 'L', 
             'X0': 'X', 'X1': 'X', 'X2': 'X'},
            [('M0.a', 'L0.a'), ('M0.b', 'X0.a'), 
             ('M0.c', 'X1.a'), ('M0.d', 'X2.a')]
        ),
        4: Assembly(  # cis-ML2X2
            {'M0': 'M', 'L0': 'L', 'L1': 'L', 
             'X0': 'X', 'X1': 'X'},
            [('M0.a', 'L0.a'), ('M0.b', 'L1.a'), 
             ('M0.c', 'X0.a'), ('M0.d', 'X1.a')]
        ),
        5: Assembly(  # trans-ML2X2
            {'M0': 'M', 'L0': 'L', 'L1': 'L', 
             'X0': 'X', 'X1': 'X'},
            [('M0.a', 'L0.a'), ('M0.b', 'X0.a'), 
             ('M0.c', 'L1.a'), ('M0.d', 'X1.a')]
        ),
    }

    expected: dict[str, Reaction] = {
        'MX4 + L -> MLX3 + X': InterReaction(
            init_assem_id=0, entering_assem_id=1,
            product_assem_id=3, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=8,
        ),
        'MLX3 + L -> cis-ML2X2 + X': InterReaction(
            init_assem_id=3, entering_assem_id=1,
            product_assem_id=4, leaving_assem_id=2,
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
        'MLX3 + L -> trans-ML2X2 + X': InterReaction(
            init_assem_id=3, entering_assem_id=1,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.c', leaving_bs='X1.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
    }

    result = explore_reactions(
        id_to_assembly,
        metal_kind='M', leaving_kind='X', entering_kind='L',
        component_structures=component_structures
    )
    assert are_equivalent_reaction_sets(
        result, list(expected.values()),
        id_to_assembly=id_to_assembly,
        component_structures=component_structures
    ), "The reaction sets with auxiliary edges are not equivalent."


def test_intra_reactions():
    component_structures = {
        'M': Component(  # component with auxiliary edges
            ['a', 'b', 'c', 'd'], 
            [('a', 'b', 'cis'), ('b', 'c', 'cis'), 
             ('c', 'd', 'cis'), ('d', 'a', 'cis')]),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
    }

    id_to_assembly = {
        0: Assembly({'X0': 'X'}),  # X
        # trans-M2L2X5
        1: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 
             'X0': 'X', 'X1': 'X', 'X2': 'X', 'X3': 'X', 'X4': 'X'},
            [('M0.a', 'L0.a'), ('M0.b', 'X0.a'), 
             ('M0.c', 'L1.a'), ('M0.d', 'X1.a'),
             ('M1.a', 'L0.b'), ('M1.b', 'X2.a'),
             ('M1.c', 'X3.a'), ('M1.d', 'X4.a')]
        ),
        # cis-M2L2X5
        2: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 
             'X0': 'X', 'X1': 'X', 'X2': 'X', 'X3': 'X', 'X4': 'X'},
            [('M0.a', 'L0.a'), ('M0.b', 'L1.a'), 
             ('M0.c', 'X0.a'), ('M0.d', 'X1.a'),
             ('M1.a', 'L0.b'), ('M1.b', 'X2.a'),
             ('M1.c', 'X3.a'), ('M1.d', 'X4.a')]
            ),
        # trans-trans-M2L2X4-ring
        3: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 
             'X0': 'X', 'X1': 'X', 'X2': 'X', 'X3': 'X'},
            [('M0.a', 'L0.a'), ('M0.b', 'X0.a'), 
             ('M0.c', 'L1.a'), ('M0.d', 'X1.a'),
             ('M1.a', 'L0.b'), ('M1.b', 'X2.a'),
             ('M1.c', 'L1.b'), ('M1.d', 'X3.a')]
            ),
        # cis-cis-M2L2X4-ring
        4: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 
             'X0': 'X', 'X1': 'X', 'X2': 'X', 'X3': 'X'},
            [('M0.a', 'L0.a'), ('M0.b', 'L1.a'), 
             ('M0.c', 'X0.a'), ('M0.d', 'X1.a'),
             ('M1.a', 'L0.b'), ('M1.b', 'L1.b'),
             ('M1.c', 'X2.a'), ('M1.d', 'X3.a')]
            ),
    }

    expected: dict[str, Reaction] = {
        'trans-M2L2X5 -> trans-trans-M2L2X4-ring + X': IntraReaction(
            init_assem_id=1, product_assem_id=3, leaving_assem_id=0,
            metal_bs='M1.c', leaving_bs='X3.a', entering_bs='L1.b',
            duplicate_count=1,
        ),
        'cis-M2L2X5 -> cis-cis-M2L2X4-ring + X': IntraReaction(
            init_assem_id=2, product_assem_id=4, leaving_assem_id=0,
            metal_bs='M1.b', leaving_bs='X2.a', entering_bs='L1.b',
            duplicate_count=2,
        ),
    }

    result = explore_reactions(
        id_to_assembly,
        metal_kind='M', leaving_kind='X', entering_kind='L',
        component_structures=component_structures
    )
    assert are_equivalent_reaction_sets(
        result, list(expected.values()),
        id_to_assembly=id_to_assembly,
        component_structures=component_structures
    ), "The reaction sets with auxiliary edges are not equivalent."


def test_verbose(capsys):
    component_structures = {
        'M': Component(['a', 'b']),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
    }

    id_to_assembly = {
        # MX2: X0(a)-(a)M0(b)-(a)X1
        0: Assembly(
            {'M0': 'M', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        ),
        1: Assembly({'L0': 'L'}),  # L: (a)L0(b)
        2: Assembly({'X0': 'X'}),  # X: (a)X0
        # MLX: (a)L0(b)-(a)M0(b)-(a)X0
        3: Assembly(
            {'M0': 'M', 'L0': 'L', 'X0': 'X'},
            [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
        ),
        # ML2: (a)L0(b)-(a)M0(b)-(a)L1(b)
        4: Assembly(
            {'M0': 'M', 'L0': 'L', 'L1': 'L'},
            [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
        ),
        # M2L2X: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)
        5: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'X0': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'L1.a')]
        ),
        # M2LX2: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)X1
        6: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'X1.a')]
        ),
        # M2L2-ring: //-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)-//
        7: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L'},
            [('M0.b', 'L0.a'), ('L0.b', 'M1.a'), ('M1.b', 'L1.a'),
             ('L1.b', 'M0.a')]
        ),
    }

    # All "X to L" reactions among the assemblies in id_to_assembly
    # including intra- and inter-molecular reactions:
    expected: dict[str, Reaction] = {
        # Intra-molecular reactions
        'M2L2X -> M2L2-ring + X': IntraReaction(
            init_assem_id=5, 
            product_assem_id=7, leaving_assem_id=2, 
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L1.b',
            duplicate_count=1,
        ),
        # Inter-molecular reactions
        'MX2 + L -> MLX + X': InterReaction(
            init_assem_id=0, entering_assem_id=1,
            product_assem_id=3, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
        'MX2 + MLX -> M2LX2 + X': InterReaction(
            init_assem_id=0, entering_assem_id=3,
            product_assem_id=6, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'MX2 + ML2 -> M2L2X + X': InterReaction(
            init_assem_id=0, entering_assem_id=4,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
        'MLX + L -> ML2 + X': InterReaction(
            init_assem_id=3, entering_assem_id=1,
            product_assem_id=4, leaving_assem_id=2,
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'MLX + MLX -> M2L2X + X': InterReaction(
            init_assem_id=3, entering_assem_id=3,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'M2LX2 + L -> M2L2X + X': InterReaction(
            init_assem_id=6, entering_assem_id=1,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
    }
    
    expected_logs = [
        'Reaction Found (0): 5 -> 7 + 2 (metal=M0.a, leaving=X0.a, entering=L1.b) (x1)',
        'Reaction Found (1): 0 + 1 -> 3 + 2 (metal=M0.a, leaving=X0.a, entering=L0.a) (x4)',
        'Reaction Found (2): 0 + 3 -> 6 + 2 (metal=M0.a, leaving=X0.a, entering=L0.a) (x2)',
        'Reaction Found (3): 0 + 4 -> 5 + 2 (metal=M0.a, leaving=X0.a, entering=L0.a) (x4)',
        'Reaction Found (4): 3 + 1 -> 4 + 2 (metal=M0.b, leaving=X0.a, entering=L0.a) (x2)',
        'Reaction Found (5): 3 + 3 -> 5 + 2 (metal=M0.b, leaving=X0.a, entering=L0.a) (x2)',
        'Reaction Found (6): 6 + 1 -> 5 + 2 (metal=M0.a, leaving=X0.a, entering=L0.a) (x4)',
    ]

    result = explore_reactions(
        id_to_assembly,
        metal_kind='M', leaving_kind='X', entering_kind='L',
        component_structures=component_structures,
        verbose=True
    )

    assert are_equivalent_reaction_sets(
        result, list(expected.values()),
        id_to_assembly=id_to_assembly,
        component_structures=component_structures
    ), "The reaction sets with verbose output are not equivalent."

    captured = capsys.readouterr()
    output_lines = captured.out.strip().split('\n')
    assert output_lines == expected_logs, "Verbose output does not match expected logs."


if __name__ == "__main__":
    pytest.main([__file__, '-v'])



import pytest
import yaml

from nasap_net import Assembly, Component, is_isomorphic
from nasap_net.pipelines import enumerate_assemblies_pipeline


def test_basic(tmp_path):
    # M2L3 linear: L1-M1-L2-M2-L3
    # bonds: 1, 2, 3, 4 from left to right
    INPUT_DATA = {
        'bonds': [1, 2, 3, 4],
        'bond_adjacency': {
            1: {2},
            2: {1, 3},
            3: {2, 4},
            4: {3},
        },
        'sym_ops': {
            'C2': {1: 4, 2: 3, 3: 2, 4: 1}
        },
        'component_kinds': {
            'L': Component(['a', 'b']),
            'M': Component(['a', 'b']),
            'X': Component(['a']),
        },
        'components_and_their_kinds': {
            'M1': 'M',
            'M2': 'M',
            'L1': 'L',
            'L2': 'L',
            'L3': 'L'
        },
        'bonds_and_their_binding_sites': {
            1: ['L1.b', 'M1.a'],
            2: ['M1.b', 'L2.a'],
            3: ['L2.b', 'M2.a'],
            4: ['M2.b', 'L3.a']
        },
        'capping_config': {
            'target_component_kind': 'M',
            'capping_component_kind': 'X',
            'capping_binding_site': 'a'
        }
    }

    input_path = tmp_path / "input.yaml"
    with open(input_path, 'w') as f:
        yaml.dump(INPUT_DATA, f)
    
    output_path = tmp_path / "output.yaml"
    enumerate_assemblies_pipeline(input_path, output_path)

    EXPECTED = {
        # L1--M1--X1
        0: Assembly(
            {'L1': 'L', 'M1': 'M', 'X1': 'X'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'X1.a')]),
        # L1--M1--L2
        1: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a')]),
        # X1--M1--L2--M2--X2
        2: Assembly(
            {'X1': 'X', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'X2': 'X'}, 
            [('X1.a', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'),
             ('M2.b', 'X2.a')]),
        # L1--M1--L2--M2--X1
        3: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'X1': 'X'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'),
             ('M2.b', 'X1.a')]),
        # L1--M1--L2--M2--L3
        4: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'L3': 'L'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'), 
             ('M2.b', 'L3.a')]),
        }
    
    with open(output_path, 'r') as f:
        actual_output = yaml.safe_load(f)

    conponent_kinds = INPUT_DATA['component_kinds']
    for index, assembly in EXPECTED.items():
        assert is_isomorphic(
            assembly, actual_output[index], 
            INPUT_DATA['component_kinds'])  # type: ignore


if __name__ == "__main__":
    pytest.main(['-v', __file__])



from collections.abc import Mapping

from nasap_net import Assembly, Component, InterOrIntra, InterReaction, \
    IntraReaction
from ..bindsite_equivalence import are_equivalent_binding_site_lists


def are_equivalent_reactions(
        reaction1: IntraReaction | InterReaction,
        reaction2: IntraReaction | InterReaction,
        id_to_assembly: Mapping[int, Assembly],
        component_structures: Mapping[str, Component],
        ) -> bool:
    """
    Check if two reactions are equivalent.

    Two reactions are equivalent if they meet the following conditions:
        1-1. They have the same number of reactants.
        1-2. They have the same number of products.
        2-1. They have the same initial assembly.
        2-2. They have the same entering assembly, or both are None.
        2-3. They have the same product assembly.
        2-4. They have the same leaving assembly, or both are None.
        3-1. If the reaction is intra-molecular, the trio of binding sites
            [metal, leaving, entering] must be equivalent.
        3-2. If the reaction is inter-molecular, the pair of binding sites
            [metal, leaving] of the initial assembly must be equivalent,
            and the entering binding site of the entering assembly must be
            equivalent to the entering binding site of the other reaction.
    
    Here, two binding site lists (e.g. [metal1, leaving1] and 
    [metal2, leaving2]) are equivalent if there is at least one isomorphism
    which maps each binding site in the first list to a binding site in 
    the second list. The order of the binding sites in the list DOES matter.
    
    Parameters
    ----------
    reaction1 : IntraReaction | InterReaction
        The first reaction to compare.
    reaction2 : IntraReaction | InterReaction
        The second reaction to compare.
    id_to_assembly : Mapping[int, Assembly]
        A mapping from assembly IDs to assemblies. This must not include
        duplicate assemblies since this function only uses the IDs to
        compare the assemblies, not the assembly structures.
    component_structures : Mapping[str, Component]
        A mapping from component IDs to components.

    Returns
    -------
    bool
        True if the reactions are equivalent, False otherwise.
    
    Warnings
    --------
    This function treats assemblies with different IDs as distinct, 
    even if they are structurally identical. Please ensure that the 
    assemblies in `id_to_assembly` are not duplicates.
    """
    # Condition 1: Same number of reactants and products
    if len(reaction1.reactants) != len(reaction2.reactants):
        return False
    if len(reaction1.products) != len(reaction2.products):
        return False
    
    # Condition 2: Same assemblies
    if reaction1.init_assem_id != reaction2.init_assem_id:
        return False
    if reaction1.entering_assem_id is not None:
        if reaction1.entering_assem_id != reaction2.entering_assem_id:
            return False
    if reaction1.product_assem_id != reaction2.product_assem_id:
        return False
    if reaction1.leaving_assem_id is not None:
        if reaction1.leaving_assem_id != reaction2.leaving_assem_id:
            return False

    # Condition 3: Equivalent pair/trio of binding sites
    if reaction1.inter_or_intra == InterOrIntra.INTRA:
        if not are_equivalent_binding_site_lists(
                id_to_assembly[reaction1.init_assem_id],
                (reaction1.metal_bs, reaction1.leaving_bs,
                 reaction1.entering_bs),
                (reaction2.metal_bs, reaction2.leaving_bs,
                 reaction2.entering_bs),
                component_structures
                ):
            return False
    elif reaction1.inter_or_intra == InterOrIntra.INTER:
        assert reaction1.entering_assem_id is not None
        # Check for initial assembly
        if not are_equivalent_binding_site_lists(
                id_to_assembly[reaction1.init_assem_id],
                (reaction1.metal_bs, reaction1.leaving_bs),
                (reaction2.metal_bs, reaction2.leaving_bs),
                component_structures
                ):
            return False
        # Check for entering assembly
        if not are_equivalent_binding_site_lists(
                id_to_assembly[reaction1.entering_assem_id],
                (reaction1.entering_bs,), (reaction2.entering_bs,),
                component_structures
                ):
            return False
    
    assert reaction1.duplicate_count == reaction2.duplicate_count
    return True



from collections.abc import Mapping

from nasap_net import Assembly, Component
from nasap_net.algorithms import are_equivalent_binding_site_lists
from nasap_net.types import C
from ..models import _MLE


def _are_equivalent_mles(
        init_assembly: Assembly,
        entering_assembly: Assembly | None,
        binding_site_trio1: _MLE,
        binding_site_trio2: _MLE,
        components: Mapping[C, Component]
        ) -> bool:
    """Check if two reactions of the same assemblies are equivalent."""
    if entering_assembly is None:  # intra
        if not are_equivalent_binding_site_lists(
                init_assembly,
                (
                        binding_site_trio1.metal,
                        binding_site_trio1.leaving,
                        binding_site_trio1.entering),
                (
                        binding_site_trio2.metal,
                        binding_site_trio2.leaving,
                        binding_site_trio2.entering),
                components
                ):
            return False
        return True
    else:  # inter
        # Check for initial assembly
        if not are_equivalent_binding_site_lists(
                init_assembly,
                (binding_site_trio1.metal, binding_site_trio1.leaving),
                (binding_site_trio2.metal, binding_site_trio2.leaving),
                components
                ):
            return False
        # Check for entering assembly
        if not are_equivalent_binding_site_lists(
                entering_assembly,
                (binding_site_trio1.entering,),
                (binding_site_trio2.entering,),
                components
                ):
            return False
        return True



import pytest

from nasap_net.pipelines.bondset_enumeration import parse_sym_ops


def test_with_sym_maps():
    input_data = {
        'C2': {1: 4, 2: 3, 3: 2, 4: 1}
    }
    expected = {
        'C2': {1: 4, 2: 3, 3: 2, 4: 1}
    }
    assert parse_sym_ops(input_data) == expected


def test_with_sym_perms():
    input_data = {
        'C3': [[1, 2, 3]]
    }
    expected = {
        'C3': {1: 2, 2: 3, 3: 1}
    }
    assert parse_sym_ops(input_data) == expected


def test_with_both_sym_maps_and_sym_perms():
    input_data = {
        'C2': {1: 4, 2: 3, 3: 2, 4: 1},
        'C3': [[1, 2, 3]]
    }
    # sym_maps should be prioritized
    expected = {
        'C2': {1: 4, 2: 3, 3: 2, 4: 1},
        'C3': {1: 2, 2: 3, 3: 1}
    }
    assert parse_sym_ops(input_data) == expected


def test_with_no_sym_ops():
    input_data = {}  # type: ignore
    assert parse_sym_ops(input_data) == {}


def test_op_chain():
    input_data = {
        'op1': {1: 2, 2: 3, 3: 1},
        'op2': {1: 1, 2: 3, 3: 2},
        'op3': ['op2', 'op1']  # op2(op1(x))
    }

    expected = {
        'op1': {1: 2, 2: 3, 3: 1},
        'op2': {1: 1, 2: 3, 3: 2},
        'op3': {1: 3, 2: 2, 3: 1},  # op2(op1(x))
    }
    assert parse_sym_ops(input_data) == expected


def test_chain_of_op_chain():
    input_data = {
        'op1': {1: 2, 2: 3, 3: 1},
        'op2': {1: 1, 2: 3, 3: 2},
        'op3': ['op2', 'op1'],
        'op4': ['op1', 'op3']  # op1(op3(x)) = op1(op2(op1(x)))
    }

    expected = {
        'op1': {1: 2, 2: 3, 3: 1},
        'op2': {1: 1, 2: 3, 3: 2},
        'op3': {1: 3, 2: 2, 3: 1},  # op2(op1(x))
        'op4': {1: 1, 2: 3, 3: 2},  # op1(op3(x))
    }
    assert parse_sym_ops(input_data) == expected



if __name__ == '__main__':
    pytest.main(['-vv', __file__])


from collections.abc import Mapping
from dataclasses import dataclass
from typing import TypeVar

from nasap_net.models import Assembly
from nasap_net.models.conversion import convert_assemblies_to_light_ones
from nasap_net.types import ID
from nasap_net.yaml.lib import dump_components, dump_light_assemblies

_T = TypeVar('_T', bound=ID)

def dump(assemblies: Mapping[_T, Assembly]) -> str:
    """Dump assemblies and components into a YAML string."""
    dumped = _dump_separately(assemblies)
    return '---\n'.join([dumped.components, dumped.assemblies])


@dataclass(frozen=True)
class _Dumped:
    assemblies: str
    components: str


def _dump_separately(
        assemblies: Mapping[_T, Assembly],
        ) -> _Dumped:
    res = convert_assemblies_to_light_ones(assemblies)

    return _Dumped(
        assemblies=dump_light_assemblies(
            dict(sorted(res.light_assemblies.items()))
        ),
        components=dump_components(
            dict(sorted(res.components.items())),
        ),
    )



from collections.abc import Mapping

from nasap_net import Assembly, Component

__all__ = ['has_aux_edges']


def has_aux_edges(
        assem: Assembly,
        component_structures: Mapping[str, Component]
        ) -> bool:
    """Check if the assembly has auxiliary edges."""
    comps_with_aux_edges = {
        comp_kind for comp_kind, comp_struct in component_structures.items()
        if comp_struct.aux_edges}
    return any(
        comp_kind in comps_with_aux_edges
        for comp_kind in assem.component_kinds)



from nasap_net.models import Assembly, Component, LightAssembly
from nasap_net.models.conversion import convert_light_assemblies_to_rich_ones
from nasap_net.types import ID
from nasap_net.yaml.helper import split_yaml_documents
from nasap_net.yaml.lib import load_components, load_light_assemblies


def load(yaml_str: str) -> dict[ID, Assembly]:
    components, light_assemblies = _load_components_and_light_assemblies(
        yaml_str=yaml_str,
    )
    return convert_light_assemblies_to_rich_ones(
        light_assemblies, components
    )


def _load_components_and_light_assemblies(
        yaml_str: str,
) -> tuple[dict[str, Component], dict[ID, LightAssembly]]:
    docs = list(split_yaml_documents(yaml_str))
    if len(docs) != 2:
        raise ValueError(
            f"Expected exactly 2 YAML documents, found {len(docs)}.")
    components = load_components(docs[0])
    light_assemblies = load_light_assemblies(docs[1])
    return components, light_assemblies



from collections.abc import Mapping

from cachetools import cached

from nasap_net import Assembly, Component
from nasap_net.algorithms import isomorphisms_iter

_cache = {}  # type: ignore

def clear_cache_for_iter_self_isomorphisms():
    """Clear the cache used by `iter_self_isomorphisms_with_cache`."""
    _cache.clear()


def _cache_key(
        assembly_id: str,
        assembly: Assembly,
        component_structures: Mapping[str, Component],
        ) -> str:
    return assembly_id


@cached(cache=_cache, key=_cache_key)
def iter_self_isomorphisms_with_cache(
        assembly_id: str, assembly: Assembly,
        component_structures: Mapping[str, Component]
        ) -> list[dict[str, str]]:
    """Iterate over self-isomorphisms of an assembly.

    Parameters
    ----------
    assembly_id : str
        The ID of the assembly. Used for caching.
    assembly : Assembly
        The assembly.
    component_structures : Mapping[str, ComponentStructure]
        A mapping from component kind to its structure.

    Returns
    -------
    list[dict[str, str]]
        A list of isomorphisms.
        Each isomorphism is a mapping from node 
        (core or bindsite, e.g., 'M1.core', 'M1.a')

    Note
    -----
    The result is cached to avoid redundant computation.
    The key for caching is based on the assembly_id, 
    i.e., assembly and component_structures are not used.
    Make sure to provide the same assembly_id for the same assembly, 
    and not to change the component_structures between calls.
    """
    return list(isomorphisms_iter(
        assembly, assembly, component_structures))
